import os
from vipy.util import remkdir, isstring, quietprint, filetail, dirlist, imlist
from vipy.image import ImageDetection
from vipy.geometry import BoundingBox
import numpy as np


URL = 'http://vis-www.cs.umass.edu/lfw/lfw.tgz'
URL_DEEPFUNNEL = 'http://vis-www.cs.umass.edu/lfw/lfw-deepfunneled.tgz'
URL_FUNNEL = 'http://vis-www.cs.umass.edu/lfw/lfw-funneled.tgz'
URL_LFWA = 'http://www.openu.ac.il/home/hassner/data/lfwa/lfwa.tar.gz'
URL_NAMES = 'http://vis-www.cs.umass.edu/lfw/lfw-names.txt'
URL_PAIRS_DEV_TRAIN = 'http://vis-www.cs.umass.edu/lfw/pairsDevTrain.txt'
URL_PAIRS_DEV_TEST = 'http://vis-www.cs.umass.edu/lfw/pairsDevTest.txt'
URL_PAIRS_VIEW2 = 'http://vis-www.cs.umass.edu/lfw/pairs.txt'

class LFW(object):
    def __init__(self, datadir, cache_root, funneled=None, bbox_size=(125,150)):
        self.datadir = datadir
        if funneled == None:
            self.lfwdir = os.path.join(self.datadir, 'lfw')
            self.lfwpairsdir = self.lfwdir
        elif funneled == 'funneled':
            self.lfwdir = os.path.join(self.datadir, 'lfw_funneled', 'lfw_funneled')
            self.lfwpairsdir = os.path.join(self.datadir, 'lfw_funneled')
        elif funneled == 'deepfunneled':
            self.lfwdir = os.path.join(self.datadir, 'lfw_deepfunneled', 'lfw-deepfunneled')
            self.lfwpairsdir = os.path.join(self.datadir, 'lfw_deepfunneled')
        self.chipdir_base = os.path.join(cache_root, "pvr", "chips")
        self.frontdir_base = os.path.join(cache_root, "pvr", "frontalized", "frontalized")
        self.frontdir_base = os.path.join(cache_root, "pvr", "frontalized", "frontalized_raw")
        self.metadir_base = os.path.join(cache_root, "pvr", "frontalized", "meta")
        
        if not os.path.isdir(os.path.join(self.lfwdir, 'AJ_Cook')):
            raise ValueError('Download LFW dataset manually and unpack to to "%s" ' % self.lfwdir)
        if not os.path.isfile(os.path.join(self.lfwpairsdir, 'pairsDevTrain.txt')):
            raise ValueError('Download LFW pairsDevTrain.txt manually and save in "%s"" ' % self.lfwpairsdir)
        if not os.path.isfile(os.path.join(self.lfwpairsdir, 'pairsDevTest.txt')):
            raise ValueError('Download LFW pairsDevTest.txt manually and save in "%s"" ' % self.lfwpairsdir)
        if not os.path.isfile(os.path.join(self.lfwpairsdir, 'pairs.txt')):
            raise ValueError('Download LFW pairs.txt manually and save in "%s"" ' % self.lfwpairsdir)
        
        if bbox_size is None:
            self.f_bbox = lambda im: None
        elif bbox_size == "viola_jones":
            raise Exception("Not implemented")
        elif bbox_size == "dlib":
            raise Exception("Not implemented")
        else:
            try:
                width,height = bbox_size
            except:
                width, height = map(int, bbox_size.split('x'))
            self.f_bbox = lambda im: BoundingBox(xmin=(im.width()/2-width/2), ymin=(im.height()/2-height/2),
                                                    width=width, height=height)
            self.bbox = BoundingBox(centroid=(125,125), width=width, height=height)

    def __repr__(self):
        return str('<viset.lfw: %s>' % self.lfwdir)

    def subjects(self):
        """List of all subject names"""
        return [filetail(d) for d in dirlist(self.lfwdir) if filetail(d) != 'lfw']

    def _subject_images(self, subject):
        """List of Images of a subject"""
        fnames = imlist(os.path.join(self.lfwdir,subject))
        ks = [fname[-8:-4] for fname in fnames]     # filename is '.../<category>_<4-digit k>.jpg', so extract just the 4-digit k's
        return [self._ImageDetectionFromTuple(self._makeParsedTuple(subject, k)) for k in ks]

    def view_subjects(self):
        """List of all Images of all subjects"""
        quietprint('[viset.lfw][LFW.view_subjects] Traversing %s recursively for images'%self.lfwdir, verbosity=2)
        return [self._subject_images(subject) for subject in self.subjects()]
    
    def _makeParsedTuple(self, category, k):
        sighting_id = '%s_%04d'%(category, int(k))
        return category, os.path.join(self.lfwdir, category, sighting_id+'.jpg'), sighting_id

    def _ImageDetectionFromTuple(self, p):
        im = ImageDetection(filename=p[1], category=p[0], bbox=self.bbox.clone(), attributes={'SIGHTING_ID': p[2], 'TEMPLATE_ID': p[0]})
        #im.boundingbox(bbox=self.f_bbox(im))
        return im

    def _parseLine(self, line):
        row = line.split()
        if len(row) == 3:
            category, k_probe, k_gallery = row
            return self._makeParsedTuple(category, k_probe), \
                   self._makeParsedTuple(category, k_gallery)
        elif len(row) == 4:
            lbl_probe, k_probe, lbl_gallery, k_gallery = row
            return self._makeParsedTuple(lbl_probe,   k_probe), \
                   self._makeParsedTuple(lbl_gallery, k_gallery)
        else: raise ValueError(line)

    def _parse_view1_protocol2(self):
        """ View 1 (development training/testing split) - Protocol 2 (image restricted verification) """
        infiles = [os.path.join(self.lfwpairsdir, 'pairsDevTrain.txt'), os.path.join(self.lfwpairsdir, 'pairsDevTest.txt')]
        outlists = [[], []]
        for (infile, outlist) in zip(infiles, outlists):
            with open(infile, 'r') as f_read:
                expected_len = int(f_read.readline())
                outlist.extend( self._parseLine(line) for line in f_read )
                assert len(outlist) == expected_len*2
        (trainlist, testlist) = outlists;
        return (trainlist, testlist)

    def _parse_view2_protocol2(self):                                    
        """ View 2 (reporting 10-fold cross validation) - Protocol 2 (image restricted verification)"""
        infile = os.path.join(self.lfwpairsdir, 'pairs.txt')
        with open(infile, 'r') as f_read:
            expected_splits, expected_split_lens = map(int, f_read.readline().split())
            splits = [[self._parseLine(f_read.readline()) for _ in xrange(expected_split_lens*2)] for _ in xrange(expected_splits)]

            # check that parsing went well
            assert len(f_read.readlines()) == 0 # no extra lines
            assert len(splits) == expected_splits
            for s in splits: assert len(s) == expected_split_lens*2
        # Done!
        return (splits)

    def view1_protocol2(self):
        """Face verification development splits (view1), image restricted protocol with no outside data (protocol 2), See also:  http://vis-www.cs.umass.edu/lfw/lfw_update.pdf"""        
        (trainlist, testlist) = self._parse_view1_protocol2()
        imtestlist = [(self._ImageDetectionFromTuple(p),
                       self._ImageDetectionFromTuple(q)) for (p,q) in testlist]
        imtrainlist = [(self._ImageDetectionFromTuple(p),
                        self._ImageDetectionFromTuple(q)) for (p,q) in trainlist]
        return [imtrainlist, imtestlist]

    def view1_protocol2_train(self):
        """Face verification development splits (view1), image restricted protocol with no outside data (protocol 2), See also:  http://vis-www.cs.umass.edu/lfw/lfw_update.pdf"""        
        trainlist,_ = self._parse_view1_protocol2()
        imtrainlist = [(self._ImageDetectionFromTuple(p),
                        self._ImageDetectionFromTuple(q)) for (p,q) in trainlist]
        return (imtrainlist,)
    
    def view2_protocol2(self):
        """Face verification 10-fold training/testing splits (view 2), image restricted protocol with no outside data (protocol 2), See also:  http://vis-www.cs.umass.edu/lfw/lfw_update.pdf"""        
        splitlist = self._parse_view2_protocol2()
        return [[(self._ImageDetectionFromTuple(p),
                  self._ImageDetectionFromTuple(q)) for (p,q) in s] for s in splitlist]

    def get_frontalized_dir(self, im):
        raise Exception("LFW.get_frontalized_dir() shouldn't be used")
        frontdir = os.path.join(self.frontdir_base, im.category())
        #bobo.util.remkdir(frontdir)
        return frontdir

    def get_frontalized_raw_dir(self, im):
        raise Exception("LFW.get_frontalized_raw_dir() shouldn't be used")
        frontrawdir = os.path.join(self.frontrawdir_base, im.category())
        #bobo.util.remkdir(frontrawdir)
        return frontrawdir

    def get_meta_dir(self, im):
        raise Exception("LFW.get_meta_dir() shouldn't be used")
        metadir = os.path.join(self.metadir_base, im.category())
        #bobo.util.remkdir(metadir)
        return metadir

    def get_chip_path(self, im):
        """
        Get the path of where the chip will be saved.

        Also makes sure the directry exists.

        Assumes im is from the LFW dataset, of course.
        """
        raise Exception("LFW.get_chip_pah() shouldn't be used")
        chipdir = os.path.join(self.chipdir_base, im.category())
        #bobo.util.remkdir(chipdir)
        return os.path.join(chipdir, '%s.png'%im.attributes['SIGHTING_ID'])

    def view(self, view_str):
        if view_str == 'subjects':
            views = self.view_subjects()
        elif view_str == 'view1_protocol2':
            views = self.view1_protocol2()
        elif view_str == 'view1_protocol2_train':
            views = self.view1_protocol2_train()
        elif view_str == 'view2_protocol2':
            views = self.view2_protocol2()
        elif view_str == 'view2_protocol2_debug':
            views = self.view2_protocol2()
            del views[:8]
            del views[0][:551]
        else:
            raise ValueError('Unknown view "%s"'%args.view)
        return views

    def flatview_pairs(self, split):
        def sid(im): return im.attributes['SIGHTING_ID']
        def cat(im): return im.category()
        id_inds = {}
        imgs = []
        pairs = []
        def ind(im):
            try: return id_inds[sid(im)]
            except:
                ind = len(imgs)
                id_inds[sid(im)] = ind
                imgs.append(im)
                return ind
        for (a,b) in split:
            aind = ind(a)
            bind = ind(b)
            pairs.append((aind, bind, a.category() == b.category()))
        return imgs, pairs

    def sample(self, n=128):
        """Return a represenative list of 128 images"""        
        imlist = [p[0] for p in self.view1_protocol2()[0]]  # test set from view 1 for a list of images
        return np.random.choice(imlist, n)
    



import matplotlib
import sys

# if sys.platform == 'linux' or sys.platform == 'linux2':
#     try:
#         plt.switch_backend('tkagg')  # linux display
#     except:
#         matplotlib.use('Agg')
#         pass  # linux headless

from matplotlib import pyplot as plt
import matplotlib.cm as cm
import numpy as np
from vipy.util import islist, temppng, mat2gray
import cv2
import os
import sys


FIGHANDLE = {}

matplotlib.rcParams['toolbar']= 'None'
plt.ion()
plt.show()
#matplotlib.rc('text', usetex=True)  # requires latex installed

def savefig(filename=None, handle=None, pad_inches=0, dpi=None, bbox_inches='tight'):
    if handle is not None:
        plt.figure(handle)
    if filename is None:
        filename = temppng()
    plt.savefig(filename, pad_inches=pad_inches, dpi=dpi, bbox_inches=bbox_inches)
    return filename

def figure(handle=None):
    if handle is not None:
        plt.figure(handle)
    else:
        plt.figure()
    return plt

def closeall():
    global FIGHANDLE;  FIGHANDLE = {};
    return plt.close('all')

def imagesc(img, title=None, colormap=None):
    imshow(mat2gray(img), title, colormap=colormap)

def _imshow(img, title=None, colormap=None, do_updateplot=True, figure=None):
    # Short pause needed to show fig
    # See http://stackoverflow.com/questions/12670101/matplotlib-ion-function-fails-to-be-interactive
    if figure is not None:
        fig = plt.figure(figure)
        plt.clf()
    else:
        fig = plt.gcf()
        plt.clf()

    ax = plt.Axes(fig, [0., 0., 1., 1.], frameon=False)
    fig.add_axes(ax)

    plt.axis('off')
    ax.set_axis_off()
    for a in plt.gcf().axes:
        a.get_xaxis().set_visible(False)
        a.get_yaxis().set_visible(False)

    if colormap in ['grey', 'gray', 'grayscale', 'greyscale']:
        cmap=cm.Greys_r
        imh = plt.imshow(img, animated=True, interpolation='nearest', cmap=cmap)
    elif colormap is not None:
        imh = plt.imshow(np.uint8(255.0*img), animated=True, interpolation='nearest', cmap=plt.get_cmap(colormap))
    else:
        imh = plt.imshow(img, animated=True, interpolation='nearest')
    plt.autoscale(tight=True)
    plt.axis('image')
    #plt.gcf().set_tight_layout(True)  # gives warning
    if title is not None:
        plt.title(title)

    if do_updateplot:
        #plt.pause(0.00001)
        try:
            plt.gcf().canvas.flush_events()
        except:
            pass
        plt.draw()
        plt.show()

    return imh

def imshow(img, title=None, colormap=None, do_updateplot=True, figure=None):
    global FIGHANDLE
    if figure in FIGHANDLE.keys() and FIGHANDLE[figure].get_size() == img.shape[0:2]:
        # Delete all polygon and text overlays from previous drawing
        FIGHANDLE[figure].set_data(img)
        for c in plt.gca().get_children():
            if 'Text' in c.__repr__() or 'Polygon' in c.__repr__() or 'Circle' in c.__repr__() or 'Line' in c.__repr__() or  'Patch' in c.__repr__():
                try:
                    c.remove()
                except:
                    pass
        if do_updateplot:
            plt.draw()
            plt.show()
    else:
        FIGHANDLE[figure] = _imshow(img, title=title, colormap=colormap, do_updateplot=do_updateplot, figure=figure)
    pause(0.00001)  # flush
    return figure


def pause(sec=0.0001):
    plt.pause(sec)

def tracks(im, bbox, bboxcolor='green', caption=None, captioncolor='red'):
    plt.clf()
    # Short pause needed to show fig
    # See http://stackoverflow.com/questions/12670101/matplotlib-ion-function-fails-to-be-interactive
    plt.pause(0.0001)
    plt.imshow(im)
    plt.autoscale(tight=True)
    plt.axis('image')
    plt.set_cmap('gray')
    #plt.hold(True)

    if caption is None:
        caption = ['det']*len(bbox)
    elif not islist(caption):
        caption = [caption]*len(bbox)

    for (bb, cap) in zip(bbox, caption):
        # (x,y) bounding box is right and down, swap to right and up for plot
        xmin = bb[0]
        ymin = bb[1]
        xmax = bb[0]+bb[2]
        ymax = bb[1]+bb[3]
        plt.axvspan(xmin, xmax, ymin=1-np.float32(float(ymax)/float(im.shape[0])), ymax=1-np.float32(float(ymin)/float(im.shape[0])), edgecolor='g', facecolor='white', linewidth=3, fill=True, alpha=0.5, label='test')

        if cap is not None:
            plt.text(xmin, ymin, cap, bbox=dict(facecolor='white', edgecolor='g',alpha=1), fontsize=10)

    plt.draw()





def imbbox(img, xmin, ymin, xmax, ymax, bboxcaption=None, figure=None, bboxcolor='green', facecolor='white', facealpha=0.5, textcolor='black', textfacecolor='white', do_updateplot=True, do_imshow=True, colormap='gray', fontsize=10, captionoffset=(0,0)):
    """Draw bounding box"""
    if figure is not None:
        plt.figure(figure)
    else:
        plt.figure()
    figure = plt.gcf().number

    # Short pause needed to show fig
    # See http://stackoverflow.com/questions/12670101/matplotlib-ion-function-fails-to-be-interactive

    if do_imshow == True:
        #plt.clf()
        imshow(img, colormap=colormap, figure=figure, do_updateplot=False)
        #plt.hold(True)

    # (x,y) bounding box is right and down, swap to right and up for plot
    # clip_on clips anything outside the image
    plt.axvspan(xmin, xmax, ymin=1.0-np.float32(float(ymax)/float(img.shape[0])), ymax=1-np.float32(float(ymin)/float(img.shape[0])), edgecolor=bboxcolor, facecolor=facecolor, linewidth=3, fill=True, alpha=facealpha, label=None, capstyle='round', joinstyle='bevel', clip_on=True)

    if bboxcaption is not None:
        # clip_on clips anything outside the image
        plt.text(xmin+captionoffset[0], ymin+captionoffset[1], bboxcaption, color=textcolor, bbox=dict(facecolor=textfacecolor, edgecolor=textcolor, alpha=1, boxstyle='round'), fontsize=fontsize, clip_on=True)

    # Update plot only for final bbox if displaying a lot
    if do_updateplot == True:
        #plt.pause(0.00001)
        try:
            plt.gcf().canvas.flush_events()
        except:
            pass

        #plt.draw()
        #plt.show()


    pause(0.00001)
    return plt.gcf().number

def imdetection(img, imdetlist, figure=None, bboxcolor='green', colormap=None, do_caption=True, facecolor='white', facealpha=0.5, textcolor='green', textfacecolor='white', captionlist=None, fontsize=10, captionoffset=(0,0)):
    """Show bounding boxes from a list of ImageDetections on the same image, plotted in list order with optional captions """

    # try imdetlist.sort(key=(lambda im: im.attributes['RawDetectionProbability']))
    # try captionlist = [im.attributes['RawDetectionProbability'] for im in imdetlist]
    # color = {'face':'red', 'vehicle':'blue', 'person':'green'}
    # bboxcolor = [color[im.category()] for im in detlist]

    # Empty?
    if len(imdetlist) == 0:
        imshow(img, figure=figure, colormap=colormap, do_updateplot=True)
        return figure

    # Valid detections
    fig = figure
    for (k,im) in enumerate(imdetlist):
        do_imshow = True if k==0 else False  # first image only
        do_updateplot = True if k==(len(imdetlist)-1) else False  # last image only
        if do_caption and captionlist is not None:
            bboxcaption = str(captionlist[k])
        elif do_caption:
            bboxcaption = str(im.category())
        else:
            bboxcaption = None

        if islist(bboxcolor):
            bboxcolor_ = bboxcolor[k]
        else:
            bboxcolor_ = bboxcolor

        if islist(textcolor):
            textcolor_ = textcolor[k]
        else:
            textcolor_ = textcolor

        fig = imbbox(img=img, xmin=im.bbox.xmin, ymin=im.bbox.ymin, xmax=im.bbox.xmax, ymax=im.bbox.ymax, bboxcaption=bboxcaption, do_imshow=do_imshow, do_updateplot=do_updateplot, figure=fig, colormap=colormap, bboxcolor=bboxcolor_, facecolor=facecolor, facealpha=facealpha, textcolor=textcolor_, textfacecolor=textfacecolor, fontsize=fontsize, captionoffset=captionoffset)
    #plt.hold(False)
    pause(0.00001)
    return fig


def precision_recall(y_precision, x_recall, title=None):
    plt.clf()
    plt.plot(x_recall, y_precision, label='Precision-Recall curve')
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.ylim([0.0, 1.0])
    plt.xlim([0.0, 1.0])
    if title is not None:
        plt.title(title)
    plt.legend(loc="lower left")
    plt.draw()



def imframe(img, fr, color='b', markersize=10, label=None, figure=None):
    if figure is not None:
        fig = plt.figure(figure)
        #plt.hold(True)
    else:
        fig = plt.figure()
        #plt.hold(True)

    figure = plt.gcf().number

    plt.pause(0.00001)
    #fig = plt.figure(np.floor(np.random.rand()*50))
    plt.clf()

    #height = float(im.shape[0])
    #width = float(im.shape[1])
    #fig = plt.gcf()
    #fig.set_size_inches(width/height, 1, forward=False)
    ax = plt.Axes(fig, [0., 0., 1., 1.], frameon=False)
    #ax.set_axis_off()
    fig.add_axes(ax)


    if img is not None:
        b = ax.imshow(img, cmap=cm.gray)
    #plt.autoscale(tight=True)
    #plt.axis('image')
    #plt.axis('off')


    plt.axis('off')
    ax.set_axis_off()
    for a in plt.gcf().axes:
        a.get_xaxis().set_visible(False)
        a.get_yaxis().set_visible(False)

    plt.autoscale(tight=True)
    #plt.hold(True)
    plt.plot(fr[:,0],fr[:,1],'%s.' % color, markersize=markersize, axes=ax)
    #plt.hold(False)
    #ax = plt.axes([0,0,1,1])
    #ax.axis('off')
    if label is not None:
        for ((x,y),lbl) in zip(fr, label):
            #plt.text(x, y, lbl, bbox=dict(facecolor='white', edgecolor='g',alpha=1))
            ax.text(x, y, lbl, color='white')

    #plt.draw()

    pause(0.00001)
    return plt


def frame(fr, im=None, color='b.', markersize=10, figure=None, caption=None):
    if figure is not None:
        fig = plt.figure(figure)
        #plt.hold(True)
    else:
        fig = plt.figure()
        plt.clf()
        #plt.hold(True)

    ax = plt.axes([0,0,1,1])
    #b = plt.imshow(im, cmap=cm.gray)
    #plt.hold(True)
    plt.plot(fr[:,0],fr[:,1],color)
    #plt.hold(False)
    plt.axis('off');
    plt.draw()
    #return plt

def show_imgdir(imgdir):
    # >>> nsd.show_imgdir("C:\\jebyrne\\penn\\datasets\\altoids")
    print('Displaying imagery from directory ' + imgdir)
    imglist = glob.glob(os.path.join(imgdir,'*.jpg'))
    for f in imglist:
        img = cv.LoadImage(f)
        thumbnail = cv.CreateMat(img.height / 8, img.width / 8, cv.CV_8UC3)
        cv.Resize(img, thumbnail)
        cv.NamedWindow('display')
        cv.ShowImage('display', thumbnail)
        cv.WaitKey(1)  # 1ms

def colorlist():
    """Return a list of named colors"""
    colorlist = [str(name) for (name, hex) in matplotlib.colors.cnames.items()]
    primarycolorlist = ['green','blue','red','cyan','orange', 'yellow','violet']
    return primarycolorlist + [c for c in colorlist if c not in primarycolorlist]



"""Annotation tools for ground truthing vision datasets"""

import urllib

import os
#import urllib2
import re
import random
import math


# https://techblog.willshouse.com/2012/01/03/most-common-user-agents/
# Updated 2015/08/12
common_user_agents = [
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/8.0.7 Safari/600.7.12',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/600.6.3 (KHTML, like Gecko) Version/8.0.6 Safari/600.6.3',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/42.0.2311.135 Safari/537.36 Edge/12.10240',
#    'Mozilla/5.0 (iPad; CPU OS 8_4 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) Version/8.0 Mobile/12H143 Safari/600.1.4',
    'Mozilla/5.0 (Windows NT 5.1; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:38.0) Gecko/20100101 Firefox/38.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.130 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/600.7.12 (KHTML, like Gecko) Version/7.1.7 Safari/537.85.16',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Ubuntu Chromium/43.0.2357.130 Chrome/43.0.2357.130 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0)',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/600.5.17 (KHTML, like Gecko) Version/8.0.5 Safari/600.5.17',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
#    'Mozilla/5.0 (iPad; CPU OS 8_3 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) Version/8.0 Mobile/12F69 Safari/600.1.4',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.2; WOW64; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.130 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.6; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11) AppleWebKit/601.1.37 (KHTML, like Gecko) Version/9.0 Safari/601.1.37',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:38.0) Gecko/20100101 Firefox/38.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_7_5) AppleWebKit/537.78.2 (KHTML, like Gecko) Version/6.1.6 Safari/537.78.2',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.130 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64; rv:31.0) Gecko/20100101 Firefox/31.0 Iceweasel/31.8.0',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.130 Safari/537.36',
    'Mozilla/5.0 (X11; Ubuntu; Linux i686; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/600.3.18 (KHTML, like Gecko) Version/8.0.3 Safari/600.3.18',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/600.6.3 (KHTML, like Gecko) Version/7.1.6 Safari/537.85.15',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_11) AppleWebKit/601.1.39 (KHTML, like Gecko) Version/9.0 Safari/601.1.39',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.107 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:31.0) Gecko/20100101 Firefox/31.0',
    'Mozilla/5.0 (Windows NT 6.2; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; Touch; rv:11.0) like Gecko',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0;  Trident/5.0)',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.130 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.8; rv:39.0) Gecko/20100101 Firefox/39.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.132 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_2) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.125 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_6_8) AppleWebKit/534.59.10 (KHTML, like Gecko) Version/5.1.9 Safari/534.59.10',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.134 Safari/537.36',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10.10; rv:38.0) Gecko/20100101 Firefox/38.0',
    'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_5) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.89 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:40.0) Gecko/20100101 Firefox/40.0',
    'Mozilla/5.0 (Windows NT 6.1; WOW64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/43.0.2357.124 Safari/537.36',
    ]


def basicgooglesearch(tag, n_imgs):
    url = 'https://www.google.com/search?q=%s&safe=off&sout=1&tbm=isch&start=%d&sa=N' 
    fulllist = []
    for k in range(20, int(n_imgs), 20):
        user_agent = random.choice(common_user_agents)
        headers = {'User-Agent':user_agent}
                     
        search_request = urllib.request.Request(url % (tag.replace(' ','+'), k), None, headers)
        search_results = urllib.request.urlopen(search_request)
        search_data = search_results.read()

        # FIXME: use gstatic.com URLs instead of .jpg
        datalist = search_data.split('http');
        imlist = [re.findall("^http[s]?://.*\.(?:jpg|gif|png)", 'http'+d) for d in datalist]
        imlist = [im[0] for im in imlist if len(im) > 0]
        imlist_clean = [im for im in imlist if im.find('File:') == -1]
        fulllist.append(imlist_clean)
    return fulllist
    
def googlesearch(tag):
    url = 'https://www.google.com/search?tbm=isch&q=%s' % tag.replace(' ','+')
    user_agent = random.choice(common_user_agents)
    headers = {'User-Agent':user_agent}
    search_request = urllib.request.Request(url,None,headers)
    search_results = urllib.request.urlopen(search_request)
    search_data = search_results.read()

    # FIXME: support for gstatic.com URLs
    datalist = search_data.split('http');
    imlist = [re.findall("^http[s]?://.*\.(?:jpg|gif|png)", 'http'+d) for d in datalist]
    imlist = [im[0] for im in imlist if len(im) > 0]
    imlist_clean = [im for im in imlist if im.find('File:') == -1]

    #imlist = [re.findall(ur'^http[s]?://.*gstatic.*?"[ ,]', ('http'+d).decode('utf-8'), re.UNICODE) for d in datalist]  # unicode double quote?
    raise('commented out previous line, needs debugging for python3')
    imlist = [im[0][:-2] for im in imlist if len(im) > 0]
    imlist2_clean = [im for im in imlist if im.find('File:') == -1]    
    return imlist_clean + imlist2_clean

        
    
def basic_level_categories():
    # nltk.download(), install wordnet in /Users/jebyrne/.nltk, 
    # set NLTK_DATA environment variable to /Users/jebyrne/.nltk
    from nltk.corpus import wordnet     
    nouns = []
    allowed_lexnames = ['noun.animal', 'noun.artifact', 'noun.body', 'noun.food', 'noun.object', 'noun.plant']
    for synset in list(wordnet.all_synsets('n')):
        if synset.lexname in allowed_lexnames:
            nouns.append(synset.lemmas[0].name)  
            #print synset.lemma_names  # synonyms
            #nouns.append(synset.name)      
    nouns.sort()
    print(nouns)
    print(len(nouns))


# A synset is a synonym set for cognitively equivalent words
# A synset is uniquely identified by a part of speech and an offset
# wordnet._synset_from_pos_and_offset('n', 2115096)
# wordnet.synset('dog.n.01')
# the synset is identified by name.part_of_speech.sense_index

# image-net popularity percentile is based on google searches and frequency in british national corpus

  

# http://stuvel.eu/media/flickrapi-docs/documentation/
# pip-2.7 install flickrapi
# http://www.flickr.com/services/api/misc.urls.html
# http://www.flickr.com/services/developer/api/
# 3600 queries per hour per key
# http://www.flickr.com/services/api/flickr.photos.search.html
# http://wordnet.princeton.edu/man/lexnames.5WN.html
# freebase.com
# https://pypi.python.org/pypi/pyimgur/0.3.2

# pylabelme

  



import os
import csv
from bobo.cache import Cache
from bobo.video import VideoCategoryStream
from bobo.util import remkdir, isexe, isvideo, isimg, imsavelist, imlist
import shutil

def videolist(viddir=None):
    return [os.path.join(viddir, filename) for filename in os.listdir(viddir) if (isvideo(filename) and (filename[0] != '.'))]

def frames(viddir):
    for v in videolist(viddir):
        (outdir, ext) = os.path.splitext(v)
        cmd = 'ffmpeg  -r 25 -i \'%s\' -qscale:v 2 -vf "scale=-1:240" %s/%%08d.jpg &> /dev/null' % (v, remkdir(outdir))        
        print '[bobo.viset.videodir]: exporting frames from "%s" to "%s"' % (v, outdir)
        os.system(cmd)

def clean(viddir):
    for v in videolist(viddir):
        (outdir, ext) = os.path.splitext(v)
        if os.path.isdir(outdir):
            print '[bobo.viset.videodir]: removing frame directory "%s"' % (outdir)            
            shutil.rmtree(outdir)

def framelist(viddir):
    fulllist = []
    for v in videolist(viddir):
        (framedir, ext) = os.path.splitext(v)
        if os.path.isdir(framedir):
            print '[bobo.viset.videodir]: generating image list for frame directory "%s"' % (framedir)            
            fulllist.append(imlist(framedir))
    return fulllist

def framedirlist(viddir):
    return [os.path.join(viddir, filename) for filename in os.listdir(viddir) if (os.path.isdir(os.path.join(viddir, filename)) and (filename[0] != '.'))]

        
                    



import os
from bobo.viset import caltech101
from bobo.cache import Cache

caltech101.URL = ('http://www.vision.caltech.edu/Image_Datasets/Caltech256/256_ObjectCategories.tar')
caltech101.SHA1 = '2195e9a478cf78bd23a1fe51f4dabe1c33744a1c'
caltech101.VISET = 'caltech256'
caltech101.SUBDIR = os.path.join('caltech256', '256_ObjectCategories')
caltech101.cache = Cache()

def stream():
    return caltech101.stream()

def export():
    return caltech101.export()    

    





raise ValueError('FIXME')

import os
import csv
from bobo.cache import Cache
from bobo.video import VideoCategoryStream
from bobo.util import remkdir, isexe

URLS = ['http://www.nada.kth.se/cvap/actions/walking.zip',
        'http://www.nada.kth.se/cvap/actions/jogging.zip',
        'http://www.nada.kth.se/cvap/actions/running.zip',
        'http://www.nada.kth.se/cvap/actions/boxing.zip',
        'http://www.nada.kth.se/cvap/actions/handwaving.zip',
        'http://www.nada.kth.se/cvap/actions/handclapping.zip']
SHA1 = ['a3e81537271a0ab4576591774baa38c2d97b7e3a',
        '21943bdbcef9dad106db0d74661e49eaeaa15a25',
        'da83bb313edfd4455fffdda6263696fa10d43c6f',
        'adb36ed9c29c846d44d2ba15f348bd115c951bbd',
        '0792f7cd69f7f895a205c08ac212ddaa7177e370',
        'd3b81261aa822ef63d0d1523f945bfaff27814d2']
LABELS = ['walking','jogging','running','boxing','handwaving','handclapping']
VISET = 'kthactions'


def split(csvfile=None):
    if csvfile is None:
        csvfile = export()
    trainPeople = ['person02','person03','person05','person06','person07','person08','person09','person10','person22']

    vidstream = VideoCategoryStream(csvfile, cache=Cache(subdir=VISET))
    trainset = [vid for vid in vidstream if any(substr in vid.uri for substr in trainPeople)]
    testset = [vid for vid in vidstream if any(substr not in vid.uri for substr in trainPeople)]    
    return (trainset, testset)


def export_videolist(outdir=None):
    cache = Cache(cacheroot=outdir, subdir=VISET)    
    vidlist = []
    for (idx_category, category) in enumerate(os.listdir(cache.root())):
        if os.path.isdir(os.path.join(cache.root(), category)):
            for (idx_video, filename) in enumerate(os.listdir(os.path.join(cache.root(), category))):    
                [avibase,ext] = os.path.splitext(filename)
                if ext == '.avi':
                    vidlist.append( (os.path.join(category, filename), category) )
    return vidlist


def export(outdir=None, clean=False):
    # Unpack dataset
    cache = Cache(cacheroot=outdir, subdir=VISET)        
    outfile = cache.abspath('%s.csv' % VISET)            

    if clean:
        cache.clean()
    elif os.path.isfile(outfile):
        print '[bobo.viset.kthactions]: exporting "%s"' % outfile
        return outfile
                        
    print '[bobo.viset.kthactions][WARNING]: downloads will not show percent progress since content length is unknown'
    for (url, label, sha1) in zip(URLS, LABELS, SHA1):
        cache.unpack(cache.get(url, sha1), cache.abspath(label), cleanup=False)

    # Check for frame export utility
    #if not isexe('ffmpeg'):
    #    raise IOError('[bobo.viset.kthactions]: ffmpeg not found on path')
                    
    # Video list
    with open(outfile, 'wb') as csvfile:
        f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)                   
        for (idx_category, category) in enumerate(os.listdir(cache.root())):
            if os.path.isdir(os.path.join(cache.root(), category)):
                for (idx_video, filename) in enumerate(os.listdir(os.path.join(cache.root(), category))):    
                    [avibase,ext] = os.path.splitext(filename)
                    if ext == '.avi':
                        imdir = cache.abspath(os.path.join(category, avibase))                        
                        print '[bobo.viset.kthactions]: exporting "%s" to "%s"' % (filename, imdir)
                        f.writerow([os.path.join(category, avibase, 'im_%08d.png'), category]);                                        
                        cmd = "ffmpeg -i \'%s\' %s/im_%%08d.png &> /dev/null" % (os.path.join(cache.root(), category, filename), imdir)
                        remkdir(imdir)                        
                        if os.system(cmd) != 0:
                            raise IOError('Error running ffmpeg')
    return outfile

def stream(outdir=None):
    cache = Cache(cacheroot=outdir, subdir=VISET)            
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return VideoCategoryStream(csvfile, cache=cache)
        




"""Helpers to download and extract archives"""

# Adapted for viset from original authors below
# 
# Authors: Nicolas Pinto <pinto@rowland.harvard.edu>
#          Nicolas Poilvert <poilvert@rowland.harvard.edu>
# License: BSD 3 clause


#import socket
#socket.setdefaulttimeout(15)  # do not set globally?
import urllib.request, urllib.error, urllib.parse
import base64
from urllib.request import urlopen
from os import path
import hashlib
import sys
from vipy.util import isfile
from . import archive
import os
# FIXME: Remove once bz2 is included in CentOS7 vendor baseline release
try:
    import bz2
except:
    pass


def generate_sha1(filepath):
    sha1 = hashlib.sha1()
    f = open(filepath, 'rb')
    try:
        sha1.update(f.read())
    finally:
        f.close()
    return sha1.hexdigest()

def verify_sha1(filename, sha1):
    data = open(filename, 'rb').read()
    return (sha1 == hashlib.sha1(data).hexdigest())

def verify_md5(filename, md5):
    data = open(filename, 'rb').read()
    if md5 != hashlib.md5(data).hexdigest():
        raise IOError("File '%s': invalid md5 hash! You may want to delete "
                      "this corrupted file..." % filename)


def download(url, output_filename, sha1=None, verbose=True, md5=None, timeout=None, username=None, password=None):
    """Downloads file at `url` and write it in `output_dirname`"""
    if timeout == None:
        timeout = 10

    if username is None and password is None:
        page = urlopen(url, None, timeout=timeout)  # urllib for username:password options
    else:
        request = urllib.request.Request(url)
        base64string = base64.encodestring('%s:%s' % (username, password)).replace('\n', '')
        request.add_header("Authorization", "Basic %s" % base64string)   
        page = urllib.request.urlopen(request, timeout=timeout)        
        
    page_info = page.info()

    output_file = open(output_filename, 'wb+')  # will raise IOError exception on invalid permissions

    # size of the download unit
    block_size = 2 ** 15
    dl_size = 0

    if verbose:
        print("[bobo.cache]: Downloading '%s' to '%s'" % (url, output_filename))
        
    # display  progress only if we know the length
    if 'content-length' in page_info and verbose:
        # file size in Kilobytes
        file_size = int(page_info['content-length']) / 1024.
        while True:
            buffer = page.read(block_size)
            if not buffer:
                break
            dl_size += block_size / 1024
            output_file.write(buffer)
            percent = min(100, 100. * dl_size / file_size)
            status = r"Progress: %20d kilobytes [%4.1f%%]" \
                    % (dl_size, percent)
            status = status + chr(8) * (len(status) + 1) 
            print(status, end=' ') # space instead of newline
            sys.stdout.flush()
        print('')
    else:
        while True:
            buffer = page.read(block_size)
            if not buffer:
                break
            dl_size += block_size / 1024
            output_file.write(buffer)
            #percent = min(100, 100. * dl_size / file_size)
            status = r"Progress: %20d kilobytes" \
                    % (dl_size)
            status = status + chr(8) * (len(status) + 1) 
            print(status, end=' ') # space instead of newline
            sys.stdout.flush()
        print('')
        #output_file.write(page.read())

    output_file.close()

    if sha1 is not None:
        if not verify_sha1(output_filename, sha1):
            raise IOError('invalid sha1')

    if md5 is not None:
        verify_md5(output_filename, md5)


def extract(archive_filename, output_dirname, sha1=None, verbose=True):
    """Extracts `archive_filename` in `output_dirname`.

    Supported archives:
    -------------------
    * Zip formats and equivalents: .zip, .egg, .jar
    * Tar and compressed tar formats: .tar, .tar.gz, .tgz, .tar.bz2, .tz2
    * gzip compressed files
    * non-tar .bz2
    """
    if verbose:
        print("[bobo.cache]: Extracting '%s' to '%s'" % (archive_filename, output_dirname))
    if sha1 is not None:
        if verbose:
            print(" SHA-1 verification...")
        verify_sha1(archive_filename, sha1)
    try:
        archive.extract(archive_filename, output_dirname, verbose=verbose)
    except archive.UnrecognizedArchiveFormat:
        base, ext = os.path.splitext(archive_filename)
        # Handle non-tar bz2 archives
        if 'bz2' in globals() and ext == '.bz2':
            # Nothing to do if already extracted
            if os.path.exists(base):
                pass
            else:
                with open(base, 'wb') as outputfile, bz2.BZ2File(archive_filename, 'rb') as bz2_file:
                    data=bz2_file.read()
                    outputfile.write(data)
        else:
            raise



def download_and_extract(url, output_dirname, sha1=None, verbose=True):
    """Downloads and extracts archive in `url` into `output_dirname`.

    Note that `output_dirname` has to exist and won't be created by this
    function.
    """
    archive_basename = path.basename(url)
    archive_filename = path.join(output_dirname, archive_basename)
    download(url, archive_filename, sha1=sha1, verbose=verbose)
    extract(archive_filename, output_dirname, sha1=sha1, verbose=verbose)

def download_extract_cleanup(url, output_dirname, sha1=None, verbose=True):
    download_and_extract(url, output_dirname, sha1, verbose)
    os.remove(path.join(output_dirname, path.basename(url)))

def extract_and_cleanup(archive_filename, output_dirname, sha1=None, verbose=True):
    download_and_extract(url, output_dirname, sha1, verbose)
    os.remove(path.join(output_dirname, path.basename(url)))

def cache_and_extract(url, output_dirname, sha1=None, verbose=True):
    if not isfile(path.join(output_dirname, path.basename(url))):
        download_and_extract(url, output_dirname, sha1, verbose)



import os
import csv
import bobo.util
from bobo.util import remkdir, quietprint, filetail, fileext
from bobo.image import ImageCategory
import bobo.app

class CelebrityFacesInTheWild(object):
    def __init__(self, datadir=None):
        self.datadir = bobo.app.datadir() if datadir is None else datadir
        self.imdir = os.path.join(self.datadir, 'msra_cfw')

        raise ValueError("FIXME")
        if not os.path.isfile(os.path.join(self.imdir, 'dev_urls.txt')):
            raise ValueError('Download PubFig dev_urls.txt manually and save to "%s" ' % self.imdir)
        
        
    def __repr__(self):
        return str('<viset.cfw: %s>' % self.imdir)



    def rdd(sparkContext, ignore=True, fetch=True):
        """Create a resilient distributed dataset"""
        quietprint("[bobo.viset.msra_cfw]: Creating spark RDD for viset '%s'" % VISET)        
        csvfile = os.path.join(CACHE.root(), VISET, '%s.csv' % VISET)            
        if not os.path.isfile(csvfile):
            csvfile = export()
    
        # Parse CSV into RDD stream of ImageCategory objects
        return (sparkContext.textFile(csvfile)  # RDD of textfile lines
                .map(lambda row: row.encode('utf-8').split())  # CSV to list
                .map(lambda x: ImageCategory(url=x[0], key=os.path.join(VISET, x[2], x[1]),  category=x[2], ignore=ignore, fetch=fetch)))  # List to image category 
        
    def export():
        # Output file
        outfile = os.path.join(CACHE.root(), VISET, '%s.csv' % VISET)
    
        # Write images and annotations
        categorydir = os.path.join(CACHE.root(), VISET, VISET)
        k_img = 0
        with open(outfile,'w') as csvfile:
            for (idx_category, category) in enumerate(os.listdir(categorydir)):
                label = bobo.util.tofilename(category)
                if not bobo.util.is_hiddenfile(category):
                    print '[viset.library.msra_cfw]: exporting "%s"' % label
            
                    imdir = os.path.join(categorydir, category)        
                    txtfile = os.path.join(imdir, 'info.txt')
                
                    with open(txtfile,'r') as f:
                        for line in f:
                            row = line.decode('utf-8').strip().split()
                            if len(row) == 3:
                                ext = fileext(row[2])
                                csvfile.write('%s %s %s\n' % (row[2].strip(), '%s_%07d%s' % (label, k_img, ext if ext is not None and len(ext)==4 else ''), label))
                            elif len(row) == 1:
                                ext = fileext(row[0])
                                csvfile.write('%s %s %s\n' % (row[0].strip(), '%s_%07d%s' % (label, k_img, ext if ext is not None and len(ext)==4 else ''), label))
                            else:
                                print 'skipping "%s"' % line 
                            k_img += 1

        return outfile
                
    
    



raise ValueError('FIXME')

import os
import csv
from bobo.cache import Cache
from bobo.video import VideoCategoryStream
from bobo.util import remkdir, isexe

URL = 'http://serre-lab.clps.brown.edu/wp-content/uploads/2013/10/hmdb51_org.rar'
SHA1 = None
VISET = 'hmdb'


def split(csvfile=None):
    pass


def unpack(rarfile, outdir):
    os.system('unrar e %s %s' % (rarfile, outdir))
    for (idx_category, rarfile) in enumerate(os.listdir(outdir)):
        (category, ext) = os.path.splitext(rarfile)    
        if not os.path.isdir(os.path.join(outdir,category)):
            os.mkdir(os.path.join(outdir, category))
            os.mkdir(os.path.join(outdir, category, 'export'))        
            cmd = 'unrar e %s %s' % (os.path.join(outdir,rarfile), os.path.join(outdir,category))
            os.system(cmd)
            #for (idx_video, avifile) in enumerate(os.listdir(os.path.join(outdir, category))):
            #    os.system('mv \'%s\' %s' % (avifile, '%s_%04d.avi' % (category, idx_video)))

            
def export_videolist(outdir=None):
    cache = Cache(cacheroot=outdir, subdir=VISET)    
    vidlist = []
    for (idx_category, category) in enumerate(os.listdir(cache.root())):
        if os.path.isdir(os.path.join(cache.root(), category)):
            for (idx_video, filename) in enumerate(os.listdir(os.path.join(cache.root(), category))):    
                [avibase,ext] = os.path.splitext(filename)
                if ext == '.avi':
                    vidlist.append( (os.path.join(category, filename), category) )
    return vidlist


def export(outdir=None, clean=False):
    # Unpack dataset
    cache = Cache(cacheroot=outdir, subdir=VISET)        
    if clean:
        cache.clean()
                        
    rarfile = unpack(cache.get(URL, SHA1), cache.root())

    vidlist = []
    for (idx_category, category) in enumerate(os.listdir(cache.root())):
        if os.path.isdir(os.path.join(cache.root(), category)):
            for (idx_video, filename) in enumerate(os.listdir(os.path.join(cache.root(), category))):    
                [avibase,ext] = os.path.splitext(filename)
                if ext == '.avi':
                    vidlist.append( (os.path.join(category, filename), category) )
    return vidlist
    
    
    # Check for frame export utility
    #if not isexe('ffmpeg'):
    #    raise IOError('[bobo.viset.hmdb]: ffmpeg not found on path')
                    

def stream(outdir=None):
    cache = Cache(cacheroot=outdir, subdir=VISET)            
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return VideoCategoryStream(csvfile, cache=cache)
        



def export_frames(indir):
    vidlist = videolist_small(indir)     
    imdir = os.path.join(indir, 'export')   
    if not os.path.isdir(imdir):
        os.mkdir(imdir)
    
    for (k_video, (avifile, category)) in enumerate(vidlist):
        [filebase,ext] = os.path.splitext(avifile)
        outdir = os.path.join(imdir, 'video_%04d' % k_video)
        if not os.path.isdir(outdir):
            os.mkdir(outdir)
        cmd = "ffmpeg -i \'%s\' -r 25 %s/img_%%06d.png" % (os.path.join(indir, avifile), outdir)
        print cmd
        os.system(cmd)
    
def export_category(outdir, mycategory):
    vidlist = export_videolist(outdir)    
    for (idx_video, (vidfile, category)) in enumerate(vidlist):
        if category == mycategory:
            (viddir, ext) = os.path.splitext(vidfile)            
            imdir = os.path.join(outdir, category, 'export', 'video_%03d' % idx_video)
            if not os.path.isdir(imdir):
                os.mkdir(imdir)
            cmd = "ffmpeg -i \'%s\' -r 30 %s/img_%%08d.jpg" % (os.path.join(outdir,vidfile), imdir)
            print cmd
            os.system(cmd)
        



"""viset package initialization"""

# HACK: create a new builtin function object 'keyboard' to simplify package debugging
from pdb import set_trace as keyboard
import builtins
builtins.keyboard = keyboard



import numpy as np
import cv2
import multiprocessing
import sys
from time import sleep
import signal
import matplotlib.cm as cm 
from bobo.util import temppng


# FIXME: rewrite me to remove broken multiprocessing event queues
raise ValueError('OpenCV backend not supported')


WINDOWSTATE = {'focus':None, 'windows':{}}

def _num_windows():
    return len(WINDOWSTATE['windows'])

def _handle(h=None):
    if h is None and WINDOWSTATE['focus'] is not None:
        h = WINDOWSTATE['focus']
    elif h is None: 
        h = 'Figure %d' % int(_num_windows()+1)
    elif type(h) is int:
        h = 'Figure %d' % int(h)
    elif type(h) is not str:
        raise ValueError('window handle must be a string')
    return h

def _eventloop(handle, drawqueue):
    def _sigint_handler(signum, frame):
        cv2.destroyWindow(handle)        
        sys.exit()
    signal.signal(signal.SIGINT, _sigint_handler)
    signal.signal(signal.SIGTERM, _sigint_handler)    
    
    cv2.namedWindow(handle)
    cv2.resizeWindow(handle, 320, 240)
    while cv2.waitKey(10) != 27:  # ESC
        if not drawqueue.empty():
            cv2.imshow(handle, drawqueue.get())
    cv2.destroyWindow(handle)

def _flip(im, handle=None):
    if im.ndim == 2:
        imbgr = cv2.cvtColor(im, cv2.cv.CV_GRAY2BGR)
    else:
        imbgr = im
        
    (x, drawprocess, drawqueue) = WINDOWSTATE['windows'][_handle(handle)]
    while drawqueue.empty() == False:
        sleep(0.01)  
    drawqueue.put(imbgr)
    WINDOWSTATE['windows'][_handle(handle)][0] = imbgr.copy()
    
def figure(handle=None):
    global WINDOWSTATE
    handle = _handle(handle)
    WINDOWSTATE['focus'] = handle
    
    if handle not in WINDOWSTATE['windows'].keys() or WINDOWSTATE['windows'][handle][1].is_alive() == False:
        drawqueue = multiprocessing.Queue()
        drawprocess = multiprocessing.Process(target=_eventloop, args=(handle, drawqueue))
        drawprocess.start()        
        WINDOWSTATE['windows'][handle] = [np.zeros((240,320)), drawprocess, drawqueue] 
    return handle

def close(handle=None):
    global WINDOWSTATE
    handle = _handle(handle)
    if handle in WINDOWSTATE['windows'].keys():
        (im, drawprocess, drawqueue) = WINDOWSTATE['windows'][handle]
        drawprocess.terminate()
        del WINDOWSTATE['windows'][handle]
    
def closeall():
    for h in WINDOWSTATE['windows'].iterkeys():
        (im, drawprocess, drawqueue) = WINDOWSTATE['windows'][h]
        drawprocess.terminate()
    WINDOWSTATE['windows'] = {}

def imshow(im, handle=None):    
    _flip(im, figure(handle))
    
def imbbox(im, xmin, xmax, ymin, ymax, bboxcaption=None):
    pass

def rectangle(bbox, color='green', caption=None, filled=False, linewidth=1, flip=False):
    global WINDOWSTATE
    im = WINDOWSTATE['windows'][WINDOWSTATE['focus']][0]
    cv2.rectangle(im, (bbox[0], bbox[1]), (bbox[0]+bbox[2], bbox[1]+bbox[3]), cv2.cv.Scalar(color[0],color[1],color[2])), 
    WINDOWSTATE['windows'][WINDOWSTATE['focus']][0] = im
    if flip:
        imshow(im)

def ellipse(bbox, color='green', caption=None, filled=False, linewidth=1):
    pass


def circle(center, radius, color, caption, filled=False, linewidth=1):
    global WINDOWSTATE
    im = WINDOWSTATE['windows'][WINDOWSTATE['focus']][0]
    cv2.circle(im, center, radius, cv2.cv.Scalar(color[0],color[1],color[2]))
    WINDOWSTATE['windows'][WINDOWSTATE['focus']][0] = im
    imshow(im)

def _color(c):
    if type(c) is str:
        if c == 'green':
            c = cv2.cv.Scalar(0,255,0)
        elif c == 'red':
            c = cv2.cv.Scalar(0,0,255)            
        elif c == 'blue':
            c = cv2.cv.Scalar(255,0,0)            
        elif c == 'yellow':
            c = cv2.cv.Scalar(0,255,255)            
        else:
            print 'undefined color %s' % c
            c = cv2.cv.Scalar(0,0,0)            
    else:    
        pass
    return c

def _im2bgr(im):
    if im.ndim == 2:
        im = cv2.cvtColor(im, cv2.cv.CV_GRAY2BGR)
    return im    

def frame(fr, im, color, caption):
    im = _im2bgr(im)
    c = _color(color)
    for xysr in fr:
        bbox = (xysr[0], xysr[1], 10, 10)
        s = xysr[2]
        th = xysr[3]
        R = np.mat([[np.cos(th), -np.sin(th)], [np.sin(th), np.cos(th)]])
        x = np.mat([ [-10, +10, +10, -10], [-10, -10, +10, +10] ])
        y = R*((s/10)*x)
        v = [(int(xysr[0]+y[0,0]),int(xysr[1]+y[1,0])), (int(xysr[0]+y[0,1]),int(xysr[1]+y[1,1])), (int(xysr[0]+y[0,2]),int(xysr[1]+y[1,2])), (int(xysr[0]+y[0,3]),int(xysr[1]+y[1,3]))]
        cv2.line(im, v[0], v[1], color=c, thickness=1, lineType=cv2.CV_AA) 
        cv2.line(im, v[1], v[2], color=c, thickness=1, lineType=cv2.CV_AA) 
        cv2.line(im, v[2], v[3], color=c, thickness=1, lineType=cv2.CV_AA) 
        cv2.line(im, v[3], v[0], color=c, thickness=1, lineType=cv2.CV_AA)                                     
    _flip(im, figure('frame'))

def tracks(im, bbox, bboxcolor='green', caption=None, captioncolor='red'):
    im = _im2bgr(im)
    if type(bboxcolor) not in (list, tuple):
        bboxcolorlist = [_color(bboxcolor)]*len(bbox)
    else:
        bboxcolorlist = [_color(c) for c in bboxcolor]
    if type(captioncolor) not in (list, tuple):
        captioncolorlist = [_color(captioncolor)]*len(bbox)
    else:
        captioncolorlist = [_color(c) for c in captioncolor]
    for (bb,bbcolor,cap,capcolor) in zip(bbox, bboxcolorlist, caption, captioncolorlist):
        cv2.rectangle(im, (int(bb[0]), int(bb[1])), (int(bb[0])+int(bb[2]), int(bb[1])+int(bb[3])), color=bbcolor), 
        if cap is not None:
            cv2.putText(im, cap, (int(bb[0]),int(bb[1])), fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.25, color=capcolor, thickness=1, lineType=cv2.CV_AA)
    _flip(im, figure('tracks')) # update the display                
    
def scatter(fr, im, color):
    im = _im2bgr(im)
    #c = _color(color)
    n_frame = fr.shape[1]
    cmap = cm.get_cmap('jet', n_frame) 
    rgb = np.uint8(255*cmap(np.arange(n_frame)))
    for (i,xysr) in enumerate(fr.transpose()):
        cv2.circle(im, (int(xysr[0]),int(xysr[1])), radius=1, color=cv2.cv.Scalar(int(rgb[i,0]), int(rgb[i,1]), int(rgb[i,2])))
    _flip(im, figure('scatter')) # update the display                

def text(ij, caption, color):
    global WINDOWSTATE
    im = WINDOWSTATE['windows'][WINDOWSTATE['focus']][0]
    cv2.putText(im, caption, ij, fontFace=cv2.FONT_HERSHEY_SIMPLEX, fontScale=0.25, color=_color(color), thickness=1, lineType=cv2.CV_AA)
    # getTextSize() for computing ij offsets
    WINDOWSTATE['windows'][WINDOWSTATE['focus']][0] = im
    imshow(im)
    
def savefig(handle=None, filename=None):
    handle = _handle(handle)
    if filename is None:
        filename = temppng()
    cv2.imwrite(filename, WINDOWSTATE['windows'][WINDOWSTATE['focus']][0])
    return filename
    
def opticalflow(im, uv_flow):
    # https://opencv-python-tutroals.readthedocs.org/en/latest/py_tutorials/py_video/py_lucas_kanade/py_lucas_kanade.html
    (mag, ang) = cv2.cartToPolar(uv_flow[...,0], uv_flow[...,1])
    hsv = np.zeros((im.shape[0], im.shape[1], 3))
    hsv[...,1] = 255
    hsv[...,0] = ang*180/np.pi/2
    hsv[...,2] = cv2.normalize(mag, None, 0, 255, cv2.NORM_MINMAX)
    imbgr = cv2.cvtColor(np.float32(hsv), cv2.COLOR_HSV2BGR)
    imshow(imbgr)
    # FIXME: this is broken

def sparseflow(im, ijuv_flow, maxflow=16):
    im = _im2bgr(im)
    (mag, ang) = cv2.cartToPolar(ijuv_flow[:,2], ijuv_flow[:,3])
    hsv = np.zeros( (1, ijuv_flow.shape[0], 3) )
    hsv[:,:,1:2] = 255
    hsv[:,:,0:1] = ang*180/np.pi/2
    hsv[:,:,2:3] = cv2.normalize(np.clip(mag,0,maxflow), None, 0, 255, cv2.NORM_MINMAX)
    bgr = cv2.cvtColor(np.uint8(hsv), cv2.COLOR_HSV2BGR)
    for (k, ijuv) in enumerate(ijuv_flow):
        c = cv2.cv.Scalar( int(bgr[0,k,0]), int(bgr[0,k,1]), int(bgr[0,k,2]))
        cv2.circle(im, (int(ijuv[0]),int(ijuv[1])), radius=2, color=c, thickness=-1)
    _flip(im, figure('sparse optical flow')) # update the display 

    
def disparity(disp, maxdisparity=None):
    if maxdisparity is None:
        maxdisparity = np.percentile(disp[np.nonzero(disp)], 95)  # robust colormap
    imdisp = np.mat((255*np.clip(np.float32(disp), 0, maxdisparity)) / maxdisparity, dtype=np.uint8)
    imdisp = cv2.applyColorMap(imdisp, cv2.COLORMAP_JET)
    _flip(imdisp, figure('disparity')) # update the display 
    
def impolygon(im, poly, color='green'):
    im = _im2bgr(im)
    c = _color(color)
    for p in poly:
        cv2.polylines(im, p, isClosed=True, color=c, thickness=1, lineType=cv2.CV_AA)
    _flip(im, figure('polygon')) # update the display                
    



import os
import csv
from bobo.cache import Cache
import numpy as np
from bobo.util import isfile, remkdir
import gzip, struct
from array import array
import numpy
import urlparse
import cv2
from bobo.viset.stream import ImageCategoryStream

TRAIN_IMG_URL = 'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz'
TRAIN_IMG_SHA1 =  '6c95f4b05d2bf285e1bfb0e7960c31bd3b3f8a7d'
TRAIN_LBL_URL = 'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz'
TRAIN_LBL_SHA1 = '2a80914081dc54586dbdf242f9805a6b8d2a15fc'
TEST_IMG_URL = 'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz'
TEST_IMG_SHA1 = 'c3a25af1f52dad7f726cce8cacb138654b760d48'
TEST_LBL_URL = 'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz'
TEST_LBL_SHA1 = '763e7fa3757d93b0cdec073cef058b2004252c17'
VISET = 'mnist'

cache = Cache()
    
def _labels(gzfile):
    with gzip.open(gzfile, 'rb') as file:
        magic, size = struct.unpack(">II", file.read(8))
        if magic != 2049:
            raise ValueError('Magic number mismatch, expected 2049,'
                            'got %d' % magic)
        labels = array("B", file.read())
    return labels


def _imread(gzfile, index):
    """Read MNIST encoded images, adapted from: https://github.com/sorki/python-mnist/blob/master/mnist/loader.py"""
    with gzip.open(gzfile, 'rb') as file:
        magic, size, rows, cols = struct.unpack(">IIII", file.read(16))
        if magic != 2051:
            raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)
        file.seek(index*rows*cols + 16)
        image = numpy.asarray(array("B", file.read(rows*cols)).tolist())
        return numpy.reshape(image, (rows,cols))
		

def stream():
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return ImageCategoryStream(csvfile)


def export():
    outfile = cache.abspath('mnist.csv')
    
    # Fetch data necessary to initial construction (do not gunzip!)
    train_img_file = cache.get(TRAIN_IMG_URL, key=os.path.join(VISET,'train-images-idx3-ubyte.gz'))
    train_lbl_file = cache.get(TRAIN_LBL_URL, key=os.path.join(VISET,'train-labels-idx1-ubyte.gz'))
    test_img_file = cache.get(TEST_IMG_URL, key=os.path.join(VISET,'t10k-images-idx3-ubyte.gz'))
    test_lbl_file = cache.get(TEST_LBL_URL, key=os.path.join(VISET,'t10k-labels-idx1-ubyte.gz'))
    y_train = _labels(train_lbl_file).tolist()
    y_test = _labels(test_lbl_file).tolist()
    
    # Write dataset 
    with open(outfile, 'wb') as csvfile:            
        f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)

        with gzip.open(train_img_file, 'rb') as gzfile:
            magic, size, rows, cols = struct.unpack(">IIII", gzfile.read(16))
            if magic != 2051:
                raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)

            for k in range(60000):
                im = numpy.asarray(array("B", gzfile.read(rows*cols)).tolist())
                im = numpy.reshape(im, (rows,cols))
                
                imfile = os.path.join(cache.root(), VISET, 'im_%06d.png' % k)
                cv2.imwrite(imfile, im)
                f.writerow([cache.key(imfile), str(y_train[k])])
                if (k % 100) == 0:
                    print '[bobo.viset.mnist][%d/%d]: exporting training image to "%s"' % (k, 60000, imfile)

        with gzip.open(test_img_file, 'rb') as gzfile:
            magic, size, rows, cols = struct.unpack(">IIII", gzfile.read(16))
            if magic != 2051:
                raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)
            
            for k in range(10000):
                im = numpy.asarray(array("B", gzfile.read(rows*cols)).tolist())
                im = numpy.reshape(im, (rows,cols))
                
                imfile = os.path.join(cache.root(), VISET, 'im_%06d.png' % (int(k)+60000))
                cv2.imwrite(imfile, im)
                f.writerow([cache.key(imfile), str(y_test[k])])            
                if (k % 100) == 0:            
                    print '[bobo.viset.mnist][%d/%d]: exporting testing image to "%s"' % (k, 10000, imfile)
      
    # Cleanup    
    return outfile
    



import numpy as np
import bobo.app
from bobo.util import load, save, csvlist, dirlist, imlist, readcsv, filebase, imwrite, remkdir, temppng, writecsv, filepath, fileext, filetail
from bobo.image import ImageDetection, Image
from bobo.video import VideoDetection
import os
import copy


class STRFace(object):
    def __init__(self, datadir='/proj/janus3/strface'):
        self.datadir = datadir 
        self.subdatadir = [os.path.join(self.datadir, d) for d in ['detections_02OCT15', 'detections_25SEP15', 'detections_30SEP15', 'detections_12OCT15', 'detections_16OCT15', 'detections_05NOV15', 'detections_10NOV15', 'detections_13NOV15']]
        self.csvlabelfiles = [os.path.join(self.datadir, 'turklabel', f) for f in ['Batch_2121390_batch_results.csv', 'Batch_2183593_batch_results.csv']]
        self._parsed = None
        
    def __repr__(self):
        return str('<viset.strface: %s>' % self.datadir)
    
    def _parse(self):
        """Return dictionary of subjectid to list of csvfiles for mated subjects"""
        if self._parsed is not None:
            return self._parsed
        
        csv1 = readcsv(self.csvlabelfiles[0])  # turk labeling
        csv2 = readcsv(self.csvlabelfiles[1])
        csv = csv1 + csv2

        mate = [y[2].replace('"', '') for y in [x for x in csv if 'Same' in x[5]]]
        nonmate = [y[2].replace('"', '') for y in [x for x in csv if 'Different' in x[5]]]
        uniquemate = set([y[2].replace('"', '') for y in [x for x in csv if 'Same' in x[5]]])  # 
        unique = set([y[2].replace('"', '') for y in csv])

        subjects = {}    
        for (j,indir) in enumerate(self.subdatadir):
            #print '[strface.parse][%d/%d]: importing %s' % (j+1, len(self.subdatadir), indir)
        
            files = csvlist(indir)
            for (k,f) in enumerate(files):
                subjectid = filebase(f).split('_')[1]  #            
                if subjectid not in uniquemate:
                    continue  # JUST LABELED MATED SUBJECTS
                if subjectid not in subjects.keys():
                    subjects[subjectid] = [f]
                else:
                    subjects[subjectid].append(f)

        self._parsed = subjects
        return subjects

    def takesubject(self, n, subjectid=None):
        """Randomly select n frames from dataset for a single subject"""
        takelist = []
        subjectdict = self._parse()
        SCHEMA = self.schema()
        subjectid = np.random.choice(self.subjects(), 1)[0] if subjectid is None else subjectid
        for k in range(0,n):
            csvfile = np.random.choice(subjectdict[subjectid], 1)[0]
            csv = readcsv(csvfile)
            r = csv[np.random.randint(1,len(csv))] # not including header
            im = ImageDetection(filename=os.path.join(filepath(csvfile), r[0]), category=subjectid, attributes={k:v for (k,v) in zip(SCHEMA,r)}).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6]))
            takelist.append(im)
        return takelist

        
    def take(self, n):
        """Randomly select n frames from dataset"""
        takelist = []
        subjectdict = self._parse()
        SCHEMA = self.schema()        
        for subjectid in np.random.choice(self.subjects(), n):
            csvfile = np.random.choice(subjectdict[subjectid], 1)[0]
            csv = readcsv(csvfile)
            r = csv[np.random.randint(1,len(csv))] # not including header
            im = ImageDetection(filename=os.path.join(filepath(csvfile), r[0]), category=subjectid, attributes={k:v for (k,v) in zip(SCHEMA,r)}).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6]))
            takelist.append(im)
        return takelist
    
    def subjects(self):
        return self._parse().keys()


    def schema(self):
        """Schema for CSV files"""
        return ['FILENAME', 'MEDIA_ID', 'TRACK_ID' ,'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT',
                'LeftBrowLeftCorner_X', 'LeftBrowLeftCorner_Y', 'LeftBrowCenter_X', 'LeftBrowCenter_Y', 'LeftBrowRightCorner_X', 'LeftBrowRightCorner_Y', 'RightBrowLeftCorner_X',
                'RightBrowLeftCorner_Y', 'RightBrowCenter_X', 'RightBrowCenter_Y','RightBrowRightCorner_X', 'RightBrowRightCorner_Y', 'LeftEyeLeftCorner_X', 'LeftEyeLeftCorner_Y',
                'LeftEyeCenter_X', 'LeftEyeCenter_Y', 'LeftEyeRightCorner_X', 'LeftEyeRightCorner_Y', 'RightEyeLeftCorner_X', 'RightEyeLeftCorner_Y', 'RightEyeCenter_X', 'RightEyeCenter_Y',
                'RightEyeRightCorner_X', 'RightEyeRightCorner_Y', 'LeftEar_X', 'LeftEar_Y', 'NoseLeft_X', 'NoseLeft_Y', 'NoseCenter_X', 'NoseCenter_Y', 'NoseRight_X', 'NoseRight_Y', 'RightEar_X',
                'RightEar_Y', 'MouthLeftCorner_X', 'MouthLeftCorner_Y', 'MouthCenter_X', 'MouthCenter_Y', 'MouthRightCorner_X', 'MouthRightCorner_Y', 'ChinCenter_X', 'ChinCenter_Y']

    
    def dataset(self):
        """Return a generator to iterate over dataset"""
        V = []
        subjects = self._parse()
        SCHEMA = self.schema()
        for (k, (subjectid, csvfiles)) in enumerate(subjects.iteritems()):
            #print '[strface.dataset][%d/%d]: importing "%s"' % (k+1, len(subjects), subjectid)
            for (j,f) in enumerate(csvfiles):
                csv = readcsv(f)                  
                if len(csv) > 1:                    
                    frames = [ImageDetection(filename=os.path.join(filepath(f), r[0]), category=subjectid, attributes={k:v for (k,v) in zip(SCHEMA,r)}).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6])) for r in csv[1:]]
                    yield VideoDetection(frames=frames)

        
    def rdd(self, sparkContext):
        return (sparkContext.parallelize(self._parse().items())
                .flatMap(lambda (k,v): [(k,f,readcsv(f)[1:]) for f in v])
                .map(lambda (k,f,v): VideoDetection(frames=[ImageDetection(filename=os.path.join(filepath(f), r[0]), category=k).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6])).setattribute('MEDIA_ID', r[1]).setattribute('TRACK_ID', r[2]) for r in v] if len(v)>0 else [])))
    
    def broadset(self):
        """Four images and two small videos [2,30] for 2500 subjects to match distribution of CS2 media, total of 15000 encodings"""
        V = []
        subjects = {k:v for (k,v) in self._parse().iteritems() if len(v)>=6}  # 3263 -> 2878, minimum 6 videos per subject
        subjects = {k:v for (j,(k,v)) in enumerate(subjects.iteritems()) if j<2500}  # 2971 --> 2500        
        for (k, (subjectid, csvfiles)) in enumerate(subjects.iteritems()):
            print '[strface.dataset][%d/%d]: importing "%s"' % (k+1, len(subjects), subjectid)
            np.random.permutation(csvfiles)  # randomly permute videos 
            for (j,f) in enumerate(csvfiles):
                csv = readcsv(f)                  
                if len(csv) > 1:                    
                    frames = [ImageDetection(filename=os.path.join(filepath(f), r[0]), category=subjectid).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6])).setattribute('MEDIA_ID', r[1]).setattribute('TRACK_ID', r[2]) for r in csv[1:]]
                    if j < 4:
                        V.append(VideoDetection(frames=list(np.random.choice(frames,1)))) # three single images, randomly selected from video
                    elif j < 6:
                        V.append(VideoDetection(frames=list(np.random.choice(frames, np.random.randint(min(2,len(frames)), min(30,len(frames))), replace=False))))  # single video with [2,30] frames randomly selected
                    else:
                        break
        return V

    def deepset(self):
        """Ten images and five videos [2,30] for 1000 subjects to match distribution of CS2 media, total of 15000 encodings"""
        V = []
        subjects = {k:v for (k,v) in self._parse().iteritems() if len(v)>=20}  # 3263 -> 1523, minimum 20 videos per subject
        subjects = {k:v for (j,(k,v)) in enumerate(subjects.iteritems()) if j<1000}  # 1523 --> 1000        
        for (k, (subjectid, csvfiles)) in enumerate(subjects.iteritems()):
            print '[strface.dataset][%d/%d]: importing "%s"' % (k+1, len(subjects), subjectid)
            np.random.permutation(csvfiles)  # randomly permute videos 
            for (j,f) in enumerate(csvfiles):
                csv = readcsv(f)                  
                if len(csv) > 1:                    
                    frames = [ImageDetection(filename=os.path.join(filepath(f), r[0]), category=subjectid).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6])).setattribute('MEDIA_ID', r[1]).setattribute('TRACK_ID', r[2]) for r in csv[1:]]
                    if j < 10:
                        V.append(VideoDetection(frames=list(np.random.choice(frames,1, replace=False)))) # 10 single images, each randomly selected from a video
                    elif j < 15:
                        V.append(VideoDetection(frames=list(np.random.choice(frames,  np.random.randint(min(2,len(frames)), min(30,len(frames))), replace=False))))  # 5 single video, each with [2,30] frames randomly selected from video
                    else:
                        break
        return V
            

    def deeperset(self):
        """All videos, [2,30] frames, All subjects"""
        V = []
        subjects = self._parse()
        for (k, (subjectid, csvfiles)) in enumerate(subjects.iteritems()):
            print '[strface.dataset][%d/%d]: importing "%s"' % (k+1, len(subjects), subjectid)
            for (j,f) in enumerate(csvfiles):
                csv = readcsv(f)                  
                if len(csv) > 1:                    
                    frames = [ImageDetection(filename=os.path.join(filepath(f), r[0]), category=subjectid).boundingbox(xmin=float(r[3]), ymin=float(r[4]), width=float(r[5]), height=float(r[6])).setattribute('MEDIA_ID', r[1]).setattribute('TRACK_ID', r[2]) for r in csv[1:]]
                    V.append(VideoDetection(frames=list(np.random.choice(frames,  np.random.randint(min(2,len(frames)), min(30,len(frames))), replace=False))))  # 5 single video, each with [2,30] frames randomly selected from video
        return V
    
    
def exporturls(outfile, strface=STRFace()):
    """Annotated URLs file"""
    S = {k:str(j) for (j,k) in enumerate(strface.subjects())}  # subject id -> unique index
    SCHEMA = ['URL', 'SUBJECT_ID', 'VIDEO_ID', 'FRAME_ID', 'MILLISECOND'] + strface.schema()[3:]
    with open(outfile, 'w') as f:
        f.write(','.join(SCHEMA) + '\n')        
        for (i,v) in enumerate(strface.dataset()):  # video iterator
            try:
                urlpath = '/data/vision/janus/data/videosearch/' + 'downloads%s' % (filepath(v[0].filename()).split('detections')[1]) + '.url'  # FIXME: RSYNC
                url = readcsv(urlpath)[0][0]  # video URL
            except:
                urlpath = strface.datadir + '/downloads%s' % (filepath(v[0].filename()).split('detections')[1]) + '.url'  
                url = readcsv(urlpath)[0][0]  # video URL

            if ',' in url:
                raise ValueError('url contains comma!  Escape me')
            
            print '[strface.export]: exporting "%s" from "%s"' % (v.category(), url)
            for (j,im) in enumerate(v):  # frame
                d = im.attributes
                d.update({'URL':url, 'SUBJECT_ID':S[v.category()], 'MILLISECOND':filebase(im.filename()), 'VIDEO_ID':i, 'FRAME_ID':j})
                attr = ','.join([str(d[k]) for k in SCHEMA])
                f.write(attr + '\n')                

    return outfile
        

def exportcrops(strface, sparkContext, outdir):
    """Crops using strface object and sparkcontext, write to outdir"""
    """sparkContext = pyspark.SparkContext(appName='strface', environment=os.environ) """
    remkdir(outdir)
    bobo.app.setverbosity(1)
    subject_to_index = {s:k for (k,s) in enumerate(strface.subjects())}
    rdd = strface.rdd(sparkContext)            
    print '[strface.exportcrops]: Writing crops for %d subjects' % len(subject_to_index)

    def f_saveas(im, savedir):
        try:
            bw = im.boundingbox().width()
            bh = im.boundingbox().height()
            cx = im.boundingbox().x_centroid()
            cy = im.boundingbox().y_centroid()
            im = im.boundingbox(xmin=cx-0.75*bw, ymin=cy-1.1*bh, xmax=cx+0.75*bw, ymax=cy+0.75*bh)  # from Omkar
            im = im.crop()  # crop here to image rectangle
            if im.boundingbox().width() < im.boundingbox().height():
                im = im.resize(cols=256)  # smallest dimension is 256
            else:
                im = im.resize(rows=256)
            im.saveas(os.path.join(savedir, '%s' % filetail(im.filename())))
        except:
            print 'skipping'
        
    (rdd.flatMap(lambda v: v.frames())
        .map(lambda im: (remkdir(os.path.join(outdir, '%04d' % subject_to_index[im.category()], '%02d' % int(im.attributes['MEDIA_ID'][-4:]))), im))
        .foreach(lambda (savedir, im):  f_saveas(im, savedir))) 
                


class STRFaceCrops(object):
    def __init__(self, datadir='/proj/janus3/strfacecrops'):
        self.datadir = datadir 
        self._subjects = None
        
    def __repr__(self):
        return str('<viset.strfacecrops: "%s">' % self.datadir)

    def subjects(self):
        if self._subjects is None:
            self._subjects = [filebase(d) for d in dirlist(os.path.join(self.datadir, 'images'))]
        return self._subjects # cached

    def dataset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir, 'images')):
            for sd in dirlist(d):  # video index
                for f in imlist(sd):
                    im = ImageDetection(filename=f, category=filebase(d))
                    im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                    im = im.boundingbox(dilate=0.875)  # central 224x224
                    yield im

    def fastset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir, 'images')):  # subject
            for sd in dirlist(d):  # video index
                for f in imlist(sd):  # frames
                    im = ImageDetection(filename=f, category=filebase(d))
                    #im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                    #im = im.boundingbox(dilate=0.875)  # central 224x224
                    yield im

    def videos(self, s):
        return [os.path.join(self.datadir, s, d) for d in dirlist(os.path.join(self.datadir, s))]


    def take(self, n):
        """Take one image from each video of n subjects"""
        S = np.random.choice(self.subjects(), n)
        takelist = []
        for d in dirlist(os.path.join(self.datadir, 'images')):
            for sd in dirlist(d):  # video index
                if filebase(d) in S:
                    f = np.random.choice(imlist(sd),1)[0]
                    im = ImageDetection(filename=f, category=filebase(d))                
                    #im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                    #im = im.boundingbox(dilate=0.875)  # central 224x224
                    takelist.append(im)
        return takelist

    def takesubject(self, n, subjectid=None):
        """Randomly select n frames from dataset for a single subject"""
        S = np.random.choice(self.subjects(), 1)
        takelist = []
        for d in dirlist(os.path.join(self.datadir, 'images')):
            if filebase(d) in S:
                for sd in np.random.choice(dirlist(d), n):  # video index
                    f = np.random.choice(imlist(sd),1)[0]
                    im = ImageDetection(filename=f, category=filebase(d))                
                    #im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                    #im = im.boundingbox(dilate=0.875)  # central 224x224
                    takelist.append(im)
        return takelist

    def takevideo(self, n, subjectid=None):
        """Randomly select n frames from each video for a single subject"""
        S = np.random.choice(self.subjects(), 1) if subjectid is None else [subjectid]
        takelist = []
        for d in dirlist(os.path.join(self.datadir, 'images')):
            if filebase(d) in S:
                for sd in dirlist(d):  # video index
                    imset = imlist(sd)
                    F = np.random.choice(imset,n)
                    im = [ImageDetection(filename=f, category=filebase(d)) for f in F]
                    #im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                    #im = im.boundingbox(dilate=0.875)  # central 224x224
                    takelist = takelist + im
        return takelist

    



"""Parses a dlib detection file for a dataset"""
import os
from csv import DictReader
from bobo.image import ImageDetection
from bobo.video import VideoDetection
import bobo.app
import numpy as np
import janus.visualize
from bobo.show import savefig
from bobo.util import filepath, tolist

BASE_SCHEMA = ['FILENAME', 'MEDIA_ID', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT',
               'LeftBrowLeftCorner_X', 'LeftBrowLeftCorner_Y', 'LeftBrowCenter_X', 'LeftBrowCenter_Y', 'LeftBrowRightCorner_X', 'LeftBrowRightCorner_Y',
               'RightBrowLeftCorner_X', 'RightBrowLeftCorner_Y', 'RightBrowCenter_X', 'RightBrowCenter_Y', 'RightBrowRightCorner_X', 'RightBrowRightCorner_Y',
               'LeftEyeLeftCorner_X', 'LeftEyeLeftCorner_Y', 'LeftEyeCenter_X', 'LeftEyeCenter_Y', 'LeftEyeRightCorner_X', 'LeftEyeRightCorner_Y',
               'RightEyeLeftCorner_X', 'RightEyeLeftCorner_Y', 'RightEyeCenter_X', 'RightEyeCenter_Y', 'RightEyeRightCorner_X', 'RightEyeRightCorner_Y',
               'LeftEar_X', 'LeftEar_Y', 'NoseLeft_X', 'NoseLeft_Y', 'NoseCenter_X', 'NoseCenter_Y', 'NoseRight_X', 'NoseRight_Y',
               'RightEar_X', 'RightEar_Y', 'MouthLeftCorner_X', 'MouthLeftCorner_Y', 'MouthCenter_X', 'MouthCenter_Y', 'MouthRightCorner_X', 'MouthRightCorner_Y',
               'ChinCenter_X', 'ChinCenter_Y']


TRACK_SCHEMA = BASE_SCHEMA[0:2] + ['TRACK_ID'] + BASE_SCHEMA[2:]
POSE_SCHEMA = TRACK_SCHEMA + ['PITCH', 'YAW', 'ROLL']

SCHEMAS = [BASE_SCHEMA, TRACK_SCHEMA, POSE_SCHEMA]


class DlibDetections(object):
    def __init__(self, csvfile, sparkContext=None):
        self.csvfile = csvfile
        self.datadir = filepath(csvfile)
        if not os.path.isfile(self.csvfile):
            raise ValueError("Can't find csv file at '%s'" % self.csvfile)
        self.sparkContext = sparkContext
        with open(self.csvfile, 'r') as f:
            try:
                header = next(f).strip().split(',')
                self.SCHEMA = [s for s in SCHEMAS if s == header][0]
            except IndexError:
                raise ValueError, 'Unrecognized schema [%r] in %s' % (header, self.csvfile)
            except StopIteration:
                print '[viset.dlib] Warning: csv file is empty: %s' % self.csvfile
                self.SCHEMA = BASE_SCHEMA

    def __repr__(self):
        return str('<viset.dlib: %s>' % self.csvfile)

    def rdd(self):
        self.sparkContext = bobo.app.init('viset_dlib') if self.sparkContext is None else self.sparkContext
        datadir = self.datadir

        xidx = self.SCHEMA.index('FACE_X')
        yidx = self.SCHEMA.index('FACE_Y')
        widx = self.SCHEMA.index('FACE_WIDTH')
        hidx = self.SCHEMA.index('FACE_HEIGHT')

        # Parse CSV file based on protocol, all non-detection properties go in attributes dictionary
        rdd = (self.sparkContext.textFile(self.csvfile)
                .map(lambda row: row.encode('utf-8').split(',')) # CSV to list of row elements
                .filter(lambda x: 'FILENAME' not in x[0])  # hack to ignore header
                .map(lambda x: ImageDetection(filename=os.path.join(datadir, x[0]), category='face',
                                              xmin = float(x[xidx]) if len(x[xidx]) > 0 else float('nan'),
                                              ymin = float(x[yidx]) if len(x[yidx]) > 0 else float('nan'),
                                              xmax = float(x[xidx])+float(x[widx]) if ((len(x[xidx])>0) and (len(x[widx])>0)) else float('nan'),
                                              ymax = float(x[yidx])+float(x[hidx]) if ((len(x[yidx])>0) and (len(x[hidx])>0)) else float('nan'),
                                              attributes={k:v for (k,v) in zip(self.SCHEMA,x)})))  # Parse row

        if 'TRACK_ID' in self.SCHEMA:
            rdd = (rdd.keyBy(lambda im: im.attributes['MEDIA_ID'] + im.attributes['TRACK_ID']) # MEDIA_ID+TRACK_ID is unique across videos
                      .reduceByKey(lambda a,b: tolist(a)+tolist(b))
                      .map(lambda (k,fr): VideoDetection(frames=fr, attributes={'MEDIA_ID':fr[0].attributes['MEDIA_ID'], 'TRACK_ID':fr[0].attributes['TRACK_ID']}) if len(tolist(fr))>1 else fr))

        return rdd

    def stream(self):
        datadir = self.datadir

        xidx = self.SCHEMA.index('FACE_X')
        yidx = self.SCHEMA.index('FACE_Y')
        widx = self.SCHEMA.index('FACE_WIDTH')
        hidx = self.SCHEMA.index('FACE_HEIGHT')

        with open(self.csvfile, 'r') as f:
            try:
                next(f)  #skip header
            except StopIteration:
                print '[viset.dlib] Warning: csv file is empty: %s' % self.csvfile
                return []

            imset = []
            for l in f:
                x = l.strip().split(',')
                im = ImageDetection(filename=os.path.join(datadir, x[0]), category='face',
                                    xmin = float(x[xidx]) if len(x[xidx]) > 0 else float('nan'),
                                    ymin = float(x[yidx]) if len(x[yidx]) > 0 else float('nan'),
                                    xmax = float(x[xidx])+float(x[widx]) if ((len(x[xidx])>0) and (len(x[widx])>0)) else float('nan'),
                                    ymax = float(x[yidx])+float(x[hidx]) if ((len(x[yidx])>0) and (len(x[hidx])>0)) else float('nan'),
                                    attributes={k:v for (k,v) in zip(self.SCHEMA,x)})  # Parse row
                imset.append(im)

        if 'TRACK_ID' in self.SCHEMA:
            framesets = {}
            sep = '|'
            for im in imset:
                key = im.attributes['MEDIA_ID'] + sep + im.attributes['TRACK_ID']
                if key not in framesets:
                    framesets[key] = []
                framesets[key].append(im)
            videos = []
            for k, fr in framesets.iteritems():
                media_id, track_id = k.split(sep)
                vid = VideoDetection(frames=fr, attributes={'MEDIA_ID': media_id, 'TRACK_ID': track_id})
                videos.append(vid)
            return videos

        return imset




def landmarks(im):
    """Return 21x2 frame array of landmark positions in 1-21 order, NaN if occluded"""
    lmidx = BASE_SCHEMA.index('LeftBrowLeftCorner_X')
    return np.float32(np.array([im.attributes[key] if len(im.attributes[key])>0 else np.float32('nan') for key in BASE_SCHEMA[lmidx:]])).reshape(21, 2)

def eyes_nose_chin(self, im):
    """Return 4x2 frame array of left eye, right eye nose chin"""
    fr = landmarks(im)
    return fr[[8-1, 11-1, 15-1, 21-1],:]  # left eye center, right eye center, nose center  (AFLW annotation, 1-indexed)

def show(im, figure=None):
    lm = landmarks(im)
    k_valid = np.argwhere(~np.isnan(lm).any(axis=1)).ravel().tolist()
    valid_landmarks = lm[k_valid]
    bbox = im.bbox.convexhull(valid_landmarks).maxsquare().dilate(1.1)
    imc = im.grayscale().crop(bbox).resize(rows=128)
    scale = float(128.0) / float(bbox.height())
    scaled_valid_landmarks = np.array([(scale*(x-bbox.xmin), scale*(y-bbox.ymin)) for (x,y) in valid_landmarks])
    bobo.show.imframe(imc.load(), scaled_valid_landmarks, 'b', markersize=30, label=None, figure=figure)

def montage(dlib):
    imlist = dlib.stream()
    for im in imlist:
        show(im, figure=1)
        im.filename(savefig())

    return janus.visualize.montage(imlist, 128,128, crop=False, grayscale=False)



import os
import csv
from bobo.cache import Cache
from bobo.viset.stream import ImageCategoryStream

URL = 'http://www.vision.caltech.edu/Image_Datasets/Caltech101/101_ObjectCategories.tar.gz'
SHA1 = 'b8ca4fe15bcd0921dfda882bd6052807e63b4c96'
VISET = 'caltech101'
SUBDIR = os.path.join('caltech101', '101_ObjectCategories')

cache = Cache()

def stream():
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()    
    return ImageCategoryStream(csvfile)

def export():
    key = cache.get(URL, key='%s.tgz' % VISET)
    pkgdir = cache.unpack(key, unpackto=os.path.join(cache.root(), VISET), sha1=SHA1, cleanup=True)

    # Output file
    outfile = cache.abspath('%s.csv' % VISET)    
    
    # Return json or CSV file containing dataset description    
    categorydir = os.path.join(cache.root(), SUBDIR)          

    # Write to annotation stream
    with open(outfile, 'wb') as csvfile:        
        f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        for (idx_category, category) in enumerate(os.listdir(categorydir)):
            imdir = os.path.join(categorydir, category)        
            for im in os.listdir(imdir):
                f.writerow([cache.key(os.path.join(categorydir, category, im)), category]);

    # Return CSV file
    return outfile







import os
import csv
from glob import glob
from bobo.cache import Cache
from bobo.util import remkdir, isstring, filebase, quietprint, tolist, islist, is_hiddenfile
from bobo.viset.stream import ImageStream
from bobo.image import ImageDetection
from bobo.video import VideoDetection
from janus.template import GalleryTemplate
import bobo.app

VISET = 'Janus_Gallery'

cache = Cache()

def rdd(sparkContext=None, galleryname=None, split=1):
    """Create a resilient distributed dataset from a split file"""
    if galleryname is None:
        raise ValueError('galleryname is required')
    schema = ['TEMPLATE_ID', 'SUBJECT_ID', 'FILE', 'MEDIA_ID', 'SIGHTING_ID', 'FRAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT', 'RIGHT_EYE_X', 'RIGHT_EYE_Y', 'LEFT_EYE_X', 'LEFT_EYE_Y', 'NOSE_BASE_X', 'NOSE_BASE_Y']
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    csvfile = os.path.join(visetdir, '%s.csv' % (galleryname))

    # Parse CSV file based on protocol into RDD of ImageDetection objects, all non-detection properties go in attributes dictionary
    schema = ['TEMPLATE_ID', 'SUBJECT_ID', 'FILE', 'MEDIA_ID', 'SIGHTING_ID', 'FRAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT', 'RIGHT_EYE_X', 'RIGHT_EYE_Y', 'LEFT_EYE_X', 'LEFT_EYE_Y', 'NOSE_BASE_X', 'NOSE_BASE_Y']
    sc = bobo.app.init('janus_gallery') if sparkContext is None else sparkContext
    return (sc.textFile(csvfile)
              .map(lambda row: row.encode('utf-8').split(','))  # CSV to list of row elements
              .filter(lambda x: x[0] != 'TEMPLATE_ID')  # hack to ignore first row
              .map(lambda x: ImageDetection(filename=os.path.join(visetdir, x[2]), category=x[1],
                                            xmin=float(x[6]) if len(x[6]) > 0 else float('nan'),
                                            ymin=float(x[7]) if len(x[7]) > 0 else float('nan'),
                                            xmax = float(x[6])+float(x[8]) if ((len(x[6])>0) and (len(x[8])>0)) else float('nan'),
                                            ymax = float(x[7])+float(x[9]) if ((len(x[7])>0) and (len(x[9])>0)) else float('nan'),
                                            attributes={k:v for (k,v) in zip(schema,x)})))  # Parse row into ImageDetection objects


def rdd_as_templates(sparkContext=None, galleryname=None):
    """Create a resilient distributed dataset of Gallery Templates from a split file"""
    return (rdd(sparkContext=sparkContext, galleryname=galleryname)  # RDD of ImageDetections (im)
            .keyBy(lambda m: str(m.attributes['TEMPLATE_ID'])) # keyby template id only
            .reduceByKey(lambda a,b: tolist(a)+tolist(b))  # list of images/videos for each template
            .map(lambda (k,medialist): GalleryTemplate(media=tolist(medialist))))  # construct template as medialist of images and videos assigned to template id


def rdd_templates_as_video(sparkContext=None, galleryname=None, split=1):
    """Create a resilient distributed dataset from a split file"""
    if galleryname is None:
        raise ValueError('galleryname is required')
    sparkContext = bobo.app.init('janus_gallery') if sparkContext is None else sparkContext
    schema = ['TEMPLATE_ID', 'SUBJECT_ID', 'FILE', 'MEDIA_ID', 'SIGHTING_ID', 'FRAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT', 'RIGHT_EYE_X', 'RIGHT_EYE_Y', 'LEFT_EYE_X', 'LEFT_EYE_Y', 'NOSE_BASE_X', 'NOSE_BASE_Y']
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    csvfile = os.path.join(visetdir, '%s.csv' % (galleryname))

    # Parse CSV file based on protocol, all non-detection properties go in attributes dictionary
    return (sparkContext.textFile(csvfile)
            .map(lambda row: row.encode('utf-8').split(','))  # CSV to list of row elements
            .filter(lambda x: x[0] != 'TEMPLATE_ID')  # hack to ignore first row
            .map(lambda x: ImageDetection(url=cache.key(os.path.join(visetdir, x[2])), category=x[1],
                                          xmin=float(x[6]) if len(x[6]) > 0 else float('nan'),
                                          ymin=float(x[7]) if len(x[7]) > 0 else float('nan'),
                                          xmax = float(x[6])+float(x[8]) if ((len(x[6])>0) and (len(x[8])>0)) else float('nan'),
                                          ymax = float(x[7])+float(x[9]) if ((len(x[7])>0) and (len(x[9])>0)) else float('nan'),
                                          attributes={k:v for (k,v) in zip(schema,x)}))  # Parse row into
            .keyBy(lambda im: str(im.attributes['TEMPLATE_ID']))  # Construct (TemplateID, x) tuples
            .reduceByKey(lambda a,b: tolist(a)+tolist(b))  # Construct (TemplateID, [x1,x2,...,xk]) tuples for videos
            .map(lambda (k,fr): VideoDetection(frames=tolist(fr), attributes=tolist(fr)[0].attributes)))

def stream(galleryname, split=None):
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    csvfile = os.path.join(visetdir, '%s.csv' % (galleryname))
    def parser(row):
        x = row.encode('utf-8').split(',') if not islist(row) else row  # CSV to list
        schema = ['TEMPLATE_ID', 'SUBJECT_ID', 'FILE', 'MEDIA_ID', 'SIGHTING_ID', 'FRAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT', 'RIGHT_EYE_X', 'RIGHT_EYE_Y', 'LEFT_EYE_X', 'LEFT_EYE_Y', 'NOSE_BASE_X', 'NOSE_BASE_Y']
        return (ImageDetection(url=cache.key(os.path.join(visetdir, x[2])), category=x[1],
                               xmin=float(x[6]) if len(x[6]) > 0 else float('nan'),
                               ymin=float(x[7]) if len(x[7]) > 0 else float('nan'),
                               xmax = float(x[6])+float(x[8]) if ((len(x[6])>0) and (len(x[8])>0)) else float('nan'),
                               ymax = float(x[7])+float(x[9]) if ((len(x[7])>0) and (len(x[9])>0)) else float('nan'),
                               attributes={k:v for (k,v) in zip(schema,x)}))

    return ImageStream(csvfile, parser=parser, delimiter=',', rowstart=2)

def export(galleryname):
    # Dataset downloaded?
    if not os.path.isdir(os.path.join(cache.root(), VISET)):
        raise ValueError('Download Janus Gallery dataset manually to "%s" ' % os.path.join(cache.root(), VISET, galleryname))

def visetdir(galleryname):
    """Returns location of this viset"""
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    remkdir(visetdir)
    return visetdir

def nextSightingId(galleryname):
    """Return the next value to be used for a new sighting"""
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    csvfile = os.path.join(visetdir, '%s.csv' % (galleryname))

    nextid = 0
    if os.path.exists(csvfile):
        with open(csvfile, 'r') as f:
            reader = csv.DictReader(f)
            for row in reader:
                try:
                    sightingid = int(row['SIGHTING_ID'])
                    nextid = max(nextid, (sightingid))
                except ValueError:
                    pass

    return nextid+1


def enroll(galleryname, imset):
    """Create or update gallery metadata and image cache with a list of images"""
    visetdir = os.path.join(cache.root(), VISET, galleryname)
    csvfile = os.path.join(visetdir, '%s.csv' % (galleryname))
    imgdir = os.path.join(visetdir, 'img')
    remkdir(imgdir)
    schema = ['TEMPLATE_ID', 'SUBJECT_ID', 'FILE', 'MEDIA_ID', 'SIGHTING_ID', 'FRAME', 'FACE_X', 'FACE_Y', 'FACE_WIDTH', 'FACE_HEIGHT', 'RIGHT_EYE_X', 'RIGHT_EYE_Y', 'LEFT_EYE_X', 'LEFT_EYE_Y', 'NOSE_BASE_X', 'NOSE_BASE_Y']

    sighting_id = nextSightingId(galleryname)

    if not os.path.exists(csvfile):
        with open(csvfile, 'w') as f:
            writer = csv.DictWriter(f, fieldnames=schema)
            writer.writeheader()

    with open(csvfile, 'a') as f:
        writer = csv.DictWriter(f, fieldnames=schema)
        for im in imset:
            if im.attributes is None:
                im.attributes = { 'TEMPLATE_ID': im.category(), 'SUBJECT_ID': im.category() }
            filename = os.path.join(imgdir, '%08d.png' % sighting_id)
            im.attributes['FILE'] = filename
            im.attributes['SIGHTING_ID'] = str(sighting_id)
            sighting_id += 1
            im.saveas(filename)

            if filename.startswith(visetdir):
                filename = filename[len(visetdir):].lstrip('/')
            r = {}
            for (k, v) in im.attributes.iteritems():
                if k in schema:
                    r[k] = v
            r['FILE'] = filename # Write out path relative to csv file
            writer.writerow(r)




from setuptools import setup, find_packages
setup(
    name = 'vipy',
    version = '0.2',
    packages = find_packages(),
    description = 'Visym python tools for privacy preserving computer vision',
    author = 'Jeffrey Byrne',
    author_email = 'jeff@visym.com',
    url = 'https://github.com/visym/vipy',
    download_url = 'https://github.com/visym/vipy/archive/0.2.tar.gz', 
    install_requires=[
        "numpy",  
        "scipy",
        "scikit-learn",
        "matplotlib",    
        "dill",
        "pillow==5.4.1",
    ],
    keywords = ['vision', 'learning', 'ML', 'CV'], 
    classifiers = [],
)




import importlib

BACKEND = importlib.import_module('vipy.gui.using_matplotlib')
#BACKEND = importlib.import_module('bobo.gui.using_opencv')  # segfaults with multiprocessing.  why?

def backend(using='pygame'):
    global BACKEND
    BACKEND = importlib.import_module('bobo.gui.using_%s' % using)
    if using == 'pygame':
        figure()  # need figure before opencv camera otherwise segfault. Why?
    
def using(using_='pygame'):
    backend(using_)

def pause(sec=1):
    BACKEND.pause(sec)
    
def figure(handle=None):
    return BACKEND.figure(handle)

def close(title=None):
    BACKEND.close(title)

def closeall():
    return BACKEND.closeall()
    
#def fullscreen():
#    BACKEND.fullscreen()
    
def imshow(im, handle=None, colormap=None, figure=None, do_updateplot=True):    
    return BACKEND.imshow(im, handle, colormap, figure=figure, do_updateplot=do_updateplot)

def imagesc(im, handle=None, colormap=None):    
    return BACKEND.imagesc(im, handle, colormap)
    
def imbbox(im, xmin, ymin, xmax, ymax, bboxcaption=None, colormap=None, figure=None, do_updateplot=True):    
    return BACKEND.imbbox(im, xmin, ymin, xmax, ymax, bboxcaption, colormap=colormap, figure=figure, do_updateplot=do_updateplot)
            
def rectangle(bbox, color='green', caption=None, filled=False, linewidth=1):
    BACKEND.rectangle(bbox, color, caption, filled, linewidth)

def ellipse(bbox, color='green', caption=None, filled=False, linewidth=1):
    BACKEND.ellipse(bbox, color, caption, filled, linewidth)
    
def circle(center, radius, color='green', caption=None, filled=False, linewidth=1):
    BACKEND.circle(center, radius, color, caption, filled, linewidth)
    
def boundingbox(bbox, caption, color='green'):
    rectangle(bbox, caption=caption, filled=False, linewidth=1, color=color)

def tracks(im, bbox, bboxcolor='green', caption=None, captioncolor='red'):
    BACKEND.tracks(im, bbox, bboxcolor, caption, captioncolor)
    
def frame(fr, im=None, color='b.', caption=False, markersize=10, figure=1):
    return BACKEND.frame(fr, im=im, color=color, caption=caption, markersize=markersize, figure=figure)

def imframe(img, fr, color='b', markersize=20, label=None, figure=None):
    return BACKEND.imframe(img=img, fr=fr, color=color, markersize=markersize, label=label, figure=figure)
    
def scatter(fr, im=None, color='green'):
    return BACKEND.scatter(fr, im, color)

def text(ij, caption, color):    
    BACKEND.text(ij, caption, color)

def savefig(filename=None, figure=None, pad_inches=0, bbox_inches='tight', dpi=None):
    return BACKEND.savefig(filename, figure, pad_inches=pad_inches, dpi=dpi, bbox_inches=bbox_inches)

def opticalflow(im, flow):
    return BACKEND.opticalflow(im, flow)

def sparseflow(im, flow):
    return BACKEND.sparseflow(im, flow)

def disparity(disp, maxdisparity=None):
    return BACKEND.disparity(disp, maxdisparity)

def impolygon(im, poly, color='green'):
    return BACKEND.impolygon(im, poly, color)

def imdetection(img, imdetlist, figure=None, bboxcolor='green', facecolor='white', facealpha=0.5, colormap=None, do_caption=True, captionlist=None, fontsize=10, textcolor='green', captionoffset=(0,0)):
    """Show a list of ImageDetections overlayed on img.  Image must be RGB (not BGR!) or grayscale"""
    return BACKEND.imdetection(img, imdetlist, figure=figure, bboxcolor=bboxcolor, colormap=colormap, do_caption=do_caption, facecolor=facecolor, facealpha=facealpha, captionlist=captionlist, fontsize=fontsize, textcolor=textcolor, captionoffset=captionoffset)

def colorlist():
    return BACKEND.colorlist()




raise ValueError('FIXME: update me to the new viset format')


import tables
import os
from os import path
from viset.dataset import CategorizationDetectionViset
import httplib2
from BeautifulSoup import BeautifulSoup, SoupStrainer # EXTERNAL
import urlparse
import string
import viset.util
from viset.util import quietprint
import xml.etree.ElementTree as ET

class LabelMe():
    def scrape(self, baseurl, keeptest):
        http = httplib2.Http()
        (status, response) = http.request(baseurl)
        filelist = []
        for link in BeautifulSoup(response, parseOnlyThese=SoupStrainer('a')):
            if (hasattr(link, 'href')) and (link.text not in ['Name','Last modified','Size','Description','Parent Directory']):
                dirurl = urlparse.urljoin(baseurl, link['href']) # relative link
                #        print 'Directory: ' + dirurl
                (status, response) = http.request(dirurl)
                for filelink in BeautifulSoup(response, parseOnlyThese=SoupStrainer('a')):
                    if (hasattr(filelink, 'href')) and (filelink.text not in ['Name','Last modified','Size','Description','Parent Directory']):
                        fileurl = urlparse.urljoin(dirurl, filelink['href'])
                        if keeptest(fileurl): 
                            print '[viset.labelme]: ' + fileurl              
                            filelist.append(fileurl) 
        return filelist

    def parse(xmlfile):
        tree = ET.parse(xmlfile)
        root = tree.getroot()
        # http://docs.python.org/2/library/xml.etree.elementtree.html#module-xml.etree.ElementTree
        pass
    
    def export(self, verbose=False):    
        # Create empty database
        db = CategorizationDetectionViset(self._dbname, mode='w', verbose=verbose)

        # Fetch data necessary to initial construction
        print('[viset.labelme]: scraping imagery from ' + self.IMURL)
        imlist = self.scrape(self.IMURL, viset.util.isimg)
        print('[viset.labelme]: scraping annotations from ' + self.ANNOURL)        
        annolist = self.scrape(self.ANNOURL, viset.util.isxml)    

        # Create me (image list only for now)
        imstream = db.image
        for imurl in imlist:
            imstream.write(imurl)

        # Cleanup
        db.close()
        return db.abspath()

class LabelMe3(LabelMe):
    IMURL = 'http://people.csail.mit.edu/brussell/research/LabelMe/Images/'
    ANNOURL = 'http://people.csail.mit.edu/brussell/research/LabelMe/Annotations/'
    SHA1 = None
    verbose = True
    _dbname = 'labelme3'



  ## Many annotations in polygon XML format for each image
  ## store as a finite list of empty columns for max 100 points

  ## Each annotation is scraped from xml document
  ## images are defined by URL from scraping entry

  ## add a DetectionPolygonTable 

  ## http://stackoverflow.com/questions/1080411/retrieve-links-from-web-page-using-python-and-beautiful-soup


  ## matplotlib supports polygons
  ## http://matplotlib.org/api/pyplot_api.html#matplotlib.pyplot.fill

  ## https://github.com/mpitid/pylabelme



from vipy.image import ImageCategory
from vipy.show import imdetection, colorlist, savefig
from vipy.util import tolist, quietprint, imwrite, imread, tmpjpg
from copy import deepcopy
import numpy as np
import matplotlib.transforms

class SceneDetection(ImageCategory):

    def __init__(self, filename=None, url=None, category='scene', attributes=None, objects=None, array=None):
        super(SceneDetection, self).__init__(filename=filename, url=url, attributes=attributes, category=category, array=array)   # ImageCategory class inheritance        
        self.objectlist = []
        self.filename(filename)  # override filename only        
        if filename is not None and objects is not None and len(objects) > 0:
            self.__dict__ = objects[0].__dict__.copy()  # shallow copy of all object attributes
            self.filename(filename)  # override filename only
        elif url is not None and objects is not None and len(objects)>0:
            self.__dict__ = objects[0].__dict__.copy()  # shallow copy of all object attributes
            self.url(url) # override url only
        else:
            super(SceneDetection, self).__init__(filename=filename, url=url, attributes=attributes, category=category, array=array)   # ImageCategory class inheritance                   

        if objects is not None and len(objects)>0:
            #self.__dict__ = objects[0].__dict__.copy()  # shallow copy of all object attributes
            self.objectlist = objects
        self.category(category)
    
    def __repr__(self):
        str_size = ", height=%d, width=%d, color='%s'" % (self._array.shape[0], self._array.shape[1], 'gray' if self._array.ndim==2 else 'color') if self.isloaded() else ""
        #if self.invalid():  # FIXME
        #    str_file = "filename='%s'" % self._filename
        #elif self.url() is not None:
        #    str_file = "url='%s'" % self.url()
        #else:
        #    str_file = ''            
        str_file = ''
        str_category = "%scategory='%s'" % (', ' if len(str_file)>0 else '', self._category)
        str_objects = ", objects=%d" % len(self.objectlist)
        return str('<vipy.scenedetection: %s%s%s%s>' % (str_file, str_category, str_size, str_objects))

    def __len__(self):
        return len(self.objectlist)

    def __iter__(self):        
        for im in self.objectlist:
            yield im
    
    def __getitem__(self, k):        
        return self.objectlist[k]
    
    def append(self, imdet):
        self.objectlist.append(imdet)
        return self
    
    def show(self, category=None, figure=None, do_caption=True, fontsize=10, boxalpha=0.25, captionlist=None, categoryColor=None, captionoffset=(0,0), outfile=None):
        """Show a subset of object categores in current image"""
        #quietprint('[vipy.scenedetection][%s]: displaying scene' % (self.__repr__()), verbosity=2)                                            
        valid_categories = sorted(self.categories() if category is None else tolist(category))
        valid_detections = [im for im in self.objectlist if im.category() in valid_categories]        
        if categoryColor is None:
            colors = colorlist()
            categoryColor = dict([(c, colors[k]) for (k, c) in enumerate(valid_categories)])
        detection_color = [categoryColor[im.category()] for im in valid_detections]
        imdetection(self.rgb()._array, valid_detections, bboxcolor=detection_color, textcolor=detection_color, figure=figure, do_caption=do_caption, facealpha=boxalpha, fontsize=fontsize, captionlist=captionlist, captionoffset=captionoffset)
        if outfile is not None:
            savefig(outfile, figure)
        return self

    def savefig(self, outfile=None, category=None, figure=None, do_caption=True, fontsize=10, boxalpha=0.25, captionlist=None, categoryColor=None, captionoffset=(0,0), dpi=200):
        """Show a subset of object categores in current image and save to the given file"""
        outfile = outfile if outfile is not None else tmpjpg()
        self.show(category, figure, do_caption, fontsize, boxalpha, captionlist, categoryColor, captionoffset)
        savefig(outfile, figure, dpi=dpi, bbox_inches='tight', pad_inches=0)
        return outfile

    def objects(self, objectlist=None):
        if objectlist is None:
            return self.objectlist
        else:
            s = self.clone()
            s.objectlist = objectlist
            return s

    def clone(self):
        return deepcopy(self)
    
    def filter(self, f):
        self.objectlist = [im for im in self.objectlist if f(im)]
        return self

    def asdict(self, f):
        """f(im) returns (k,v) pair"""
        return {f(im) for im in self.objectlist}

    def distinct(self, f):
        return list(set([f(im) for im in self.objectlist]))

    def categories(self):
        return self.distinct(lambda im: im.category())
    
    def aslist(self, f):
        """f(im) return (v) singleton"""
        return [f(im) for im in self.objectlist]

    def overlap(self, bb):
        return [im.boundingbox().overlap(bb) for im in self.objectlist]

    def argmax(self, f):
        return self.objectlist[np.argmax(np.array([f(im) for im in self.objectlist]))]

    def argmin(self, f):
        return self.objectlist[np.argmin(np.array([f(im) for im in self.objectlist]))]
                
    def map(self, f):
        self.objectlist = [f(im) for im in self.objectlist]
        return self
    
    def sort(self, f):
        self.objectlist.sort(key=f)
        return self

    def deal(self, f, X):
        self.objectlist = [f(im,x) for (im,x) in zip(self.objectlist, X)]
        return self





import urllib
import sys
import os
import datetime
import json
from vipy.util import remkdir, imlist, dirlist
import numpy as np
import vipy.annotation


def count(outdir):
    num_files = 0
    for d in dirlist(outdir):
        num_files += len(imlist(d))
    return num_files

    
def facebookprofilerange(fbid, numid, outdir='./imgs', cleanup=True, hierarchical=False, redownload=False):
    for x in range(fbid, fbid+numid):
        facebookprofile(x, outdir, cleanup, hierarchical, redownload)

def facebookprofile(fbid, outdir='./imgs', cleanup=True, hierarchical=False, redownload=False):
    if hierarchical:
        subdir = remkdir(os.path.join(outdir, str(int(float(fbid)/1E4)))) # 10000 images per directory
        outfile = os.path.join(subdir, '%d.jpg' % int(fbid))  # outdir/1000/10000001.jpg
    else:
        outfile = os.path.join(outdir, '%d.jpg' % int(fbid))
        
    url = "http://graph.facebook.com/picture?id=" + str(fbid) + "&width=800"
    if not os.path.exists(outfile) or redownload:
        try:
            print('[facebookprofile.download]: Downloading "%s" to "%s"' % (url, outfile))
            
            user_agent = np.random.choice(vipy.annotation.common_user_agents)
            headers = {'User-Agent':user_agent}                     
            req = urllib.request.Request(url, None, headers)
            imgfile = urllib.request.urlopen(req)
            total_size = int(imgfile.info().getheader('Content-Length').strip())
            downloaded = 0
            CHUNK = 256 * 10240
            with open(outfile, 'wb') as fp:
                while True:
                    chunk = imgfile.read(CHUNK)
                    downloaded += len(chunk)
                    #print math.floor( (downloaded / total_size) * 100 )
                    if not chunk: break
                    fp.write(chunk)
                        
            #urllib.urlretrieve(url, outfile)
            
            s = os.path.getsize(outfile)
            if cleanup and (s < 11000 or s == 10626 or s == 10491):
                print('[facebookprofile.download]: deleting invalid file "%s"' % outfile)
                os.remove(outfile)

        except (urllib.request.HTTPError, e):
            print('[fb_image.download]: Skipping "%s"' % (url))
            print("HTTP Error:",e.code , url)
            #return False
        except (urllib.request.URLError, e):
            print('[fb_image.download]: Skipping "%s"' % (url))
            print("URL Error:",e.reason , url)
            #return False
        except KeyboardInterrupt:
            raise
        except:
            print("UNKNOWN ERROR")
            #raise



raise ValueError('FIXME')




import os
import csv
from bobo.cache import Cache
from bobo.util import remkdir, isstring
from bobo.image import ImageCategoryStream

URL = 'http://pascallin.ecs.soton.ac.uk/challenges/VOC/voc2012/VOCtrainval_11-May-2012.tar'
SHA1 = None
IMDIR = 'VOCdevkit/VOC2012/JPEGImages'
IMSETDIR = 'VOCdevkit/VOC2012/ImageSets/Main'
VISET = 'pascalvoc_2012'

cache = Cache(subdir=VISET)



def stream(csvfile=None, outdir=None):
    if csvfile is None:
        csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if outdir is not None:
        cache.setroot(outdir)
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return ImageCategoryStream(csvfile, cache=cache)


def export(outdir=None):
    # Update cache
    if outdir is not None:
        cache.setroot(outdir)

    # Fetch data necessary to initial construction
    pkgdir = cache.unpack(cache.get(URL), sha1=SHA1)
    imsetdir = os.path.join(pkgdir, IMSETDIR)                              
    outfile = os.path.join(cache.root(), '%s.csv' % VISET)
    
    # Write images to database
    with open(outfile, 'wb') as csvfile:            
        f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)
        for (idx_category, imset) in enumerate(os.listdir(imsetdir)):
            (filebase,ext) = os.path.splitext(os.path.basename(imset))
            try:
                (category, set) = filebase.split('_')            
            except:
                continue
            if set == 'trainval':
                for (idx_image, line) in enumerate(open(os.path.join(imsetdir, imset), 'r')):
                    (im, label) = line.split()
                    im = im + '.jpg'
                    if int(label) == 1:
                        f.writerow([os.path.join(IMDIR, im), category])                                    

    return outfile
